import logging

from angrmanagement.ui.views.view import BaseView
from binsync.common.ui.control_panel import ControlPanel
from binsync.common.ui.qt_objects import QVBoxLayout

from .controller import AngrBinSyncController


l = logging.getLogger(__name__)


class ControlPanelView(BaseView):
    """
    The class for the window that shows changes/info to BinSync data. This includes things like
    changes to functions or structs.
    """

    def __init__(self, workspace, default_docking_position, controller, *args, **kwargs):
        super().__init__('sync', workspace, default_docking_position, *args, **kwargs)

        self.base_caption = "BinSync: Control Panel"

        self.controller: AngrBinSyncController = controller
        self.control_panel = ControlPanel(self.controller)
        self._init_widgets()

        self.width_hint = 300

    def reload(self):
        pass

    #
    # Private methods
    #

    def _init_widgets(self):
        main_layout = QVBoxLayout()
        main_layout.addWidget(self.control_panel)
        self.setLayout(main_layout)



from typing import Dict, Optional

import toml

from binsync.data.artifact import Artifact
from binsync.data.stack_variable import StackVariable

#
# Function Header Classes
#

class FunctionArgument(Artifact):
    __slots__ = (
        "last_change",
        "idx",
        "name",
        "type_str",
        "size"
    )

    def __init__(self, idx, name, type_str, size, last_change=None):
        super(FunctionArgument, self).__init__(last_change=last_change)
        self.idx = idx
        self.name = name
        self.type_str = type_str
        self.size = size

    def __str__(self):
        return f"<FuncArg: {self.type_str} {self.name}; @{self.idx}>"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def parse(cls, s):
        fa = FunctionArgument(None, None, None, None)
        fa.__setstate__(toml.loads(s))
        return fa

    def copy(self):
        return FunctionArgument(self.idx, self.name, self.type_str, self.size, last_change=self.last_change)


class FunctionHeader(Artifact):
    __slots__ = (
        "last_change",
        "name",
        "addr",
        "ret_type",
        "args"
    )

    def __init__(self, name, addr, ret_type=None, args=None, last_change=None):
        super(FunctionHeader, self).__init__(last_change=last_change)
        self.name = name
        self.addr = addr
        self.ret_type = ret_type
        self.args = args or {}

    def __str__(self):
        return f"<FuncHeader: {self.ret_type} {self.name}(args={len(self.args)}); @{hex(self.addr)}>"

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        args = {str(idx): arg.__getstate__() for idx, arg in self.args.items()} if self.args else {}

        return {
            "last_change": self.last_change,
            "name": self.name,
            "addr": self.addr,
            "ret_type": self.ret_type,
            "args": args if len(args) > 0 else None,
        }

    def __setstate__(self, state):
        self.last_change = state.get("last_change", None)
        self.name = state.get("name", None)
        self.addr = state["addr"]
        self.ret_type = state.get("ret_type", None)
        args = state.get("args", {})
        self.args = {int(idx, 16): FunctionArgument.parse(toml.dumps(arg)) for idx, arg in args.items()}

    @classmethod
    def parse(cls, s):
        loaded_s = toml.loads(s)
        if len(loaded_s) <= 0:
            return None

        fh = FunctionHeader(None, None)
        fh.__setstate__(toml.loads(s))
        return fh

    def diff(self, other, **kwargs) -> Dict:
        diff_dict = {}
        # early exit if the two do not match type
        if not isinstance(other, FunctionHeader):
            for k in ["name", "addr", "ret_type"]:
                diff_dict[k] = {
                    "before": getattr(self, k),
                    "after": None
                }

            diff_dict["args"] = {idx: arg.diff(None) for idx, arg in self.args.items()}
            return diff_dict

        # metadata
        for k in ["name", "addr", "ret_type"]:
            if getattr(self, k) == getattr(other, k):
                continue

            diff_dict[k] = {
                "before": getattr(self, k),
                "after": getattr(other, k)
            }

        # args
        diff_dict["args"] = {}
        for idx, self_arg in self.args.items():
            try:
                other_arg = other.args[idx]
            except KeyError:
                other_arg = None

            diff_dict["args"][idx] = self_arg.diff(other_arg)

        for idx, other_arg in other.args.items():
            if idx in diff_dict["args"]:
                continue

            diff_dict["args"][idx] = self.invert_diff(other_arg.diff(None))

        return diff_dict

    def copy(self):
        fh = FunctionHeader(self.name, self.addr, ret_type=self.ret_type, last_change=self.last_change)
        fh.args = {k: v.copy() for k, v in self.args.items()}
        return fh

    def nonconflict_merge(self, fh2: "FunctionHeader", **kwargs):
        fh1: "FunctionHeader" = self.copy()
        if not fh2 or not isinstance(fh2, FunctionHeader):
            return fh1

        if fh1.name is None:
            fh1.name = fh2.name

        if fh1.ret_type is None:
            fh1.ret_type = fh2.ret_type

        # header args
        for off, var in fh2.args.items():
            merge_var: FunctionArgument = fh1.args[off].copy() if off in fh1.args else var
            merge_var = merge_var.nonconflict_merge(var)
            fh1.args[off] = merge_var

        return fh1

#
# Full Function Class
#

class Function(Artifact):
    """
    The Function class describes a Function found a decompiler. There are three components to a function:
    1. Metadata
    2. Header
    3. Stack Vars

    The metadata contains info on changes and size. The header holds the return type,
    and arguments (including their types). The stack vars contain StackVariables.
    """

    __slots__ = (
        "last_change",
        "addr",
        "size",
        "header",
        "stack_vars",
    )

    def __init__(self, addr, size, header=None, stack_vars=None, last_change=None):
        super(Function, self).__init__(last_change=last_change)
        self.addr: int = addr
        self.size: int = size
        self.header: Optional[FunctionHeader] = header
        self.stack_vars: Dict[int, StackVariable] = stack_vars or {}

    def __str__(self):
        if self.header:
            return f"<Function: {self.header.ret_type} {self.name}(args={len(self.args)}); @{hex(self.addr)} " \
                   f"vars={len(self.stack_vars)} len={hex(self.size)}>"

        return f"<Function: @{hex(self.addr)} len={hex(self.size)}>"

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        header = self.header.__getstate__() if self.header else None
        stack_vars = {"%x" % offset: stack_var.__getstate__() for offset, stack_var in self.stack_vars.items()} if \
            self.stack_vars else {}

        return {
            "metadata": {
                "addr": self.addr,
                "size": self.size,
                "last_change": self.last_change
            },
            "header": header,
            "stack_vars": stack_vars if len(stack_vars) > 0 else None
        }

    def __setstate__(self, state):
        if not isinstance(state["metadata"]["addr"], int):
            raise TypeError("Unsupported type %s for addr." % type(state["metadata"]["addr"]))

        metadata, header, stack_vars = state["metadata"], state.get("header", None), state.get("stack_vars", {})

        self.addr = metadata["addr"]
        self.size = metadata["size"]
        self.last_change = metadata.get("last_change", None)

        self.header = FunctionHeader.parse(toml.dumps(header)) if header else None

        self.stack_vars = {
            int(off, 16): StackVariable.parse(toml.dumps(stack_var)) for off, stack_var in stack_vars.items()
        } if stack_vars else {}

    def diff(self, other, **kwargs) -> Dict:
        diff_dict = {}
        if not isinstance(other, Function):
            # metadata
            for k in ["addr", "size"]:
                diff_dict[k] = {
                    "before": getattr(self, k),
                    "after": None
                }

            # header
            diff_dict["header"] = self.header.diff(other.header)
            # args
            diff_dict["stack_vars"] = {off: var.diff(None) for off, var in self.stack_vars.items()}
            return diff_dict

        # metadata
        for k in ["addr", "size"]:
            if getattr(self, k) == getattr(other, k):
                continue

            diff_dict[k] = {
                "before": getattr(self, k),
                "after": getattr(other, k)
            }

        # header
        if self.header:
            diff_dict["header"] = self.header.diff(other.header)
        elif other.header:
            diff_dict["header"] = self.invert_diff(other.header.diff(None))
        else:
            diff_dict["header"] = {"before": None, "after": None}

        # stack vars
        diff_dict["stack_vars"] = {}
        for off, self_var in self.stack_vars.items():
            try:
                other_var = other.stack_vars[off]
            except KeyError:
                other_var = None

            diff_dict["stack_vars"][off] = self_var.diff(other_var)

        for off, other_var in other.stack_vars.items():
            if off in diff_dict["stack_vars"]:
                continue

            diff_dict["stack_vars"][off] = self.invert_diff(other_var.diff(None))

        return diff_dict

    def copy(self):
        func = Function(self.addr, self.size, last_change=self.last_change)
        func.header = self.header.copy() if self.header else None
        func.stack_vars = {k: v.copy() for k, v in self.stack_vars.items()}

        return func

    @classmethod
    def parse(cls, s):
        func = Function(None, None)
        func.__setstate__(s)
        return func

    @classmethod
    def load(cls, func_toml):
        f = Function(None, None)
        f.__setstate__(func_toml)
        return f

    def nonconflict_merge(self, func2: "Artifact", **kwargs):
        func1: "Function" = self.copy()

        if not func2 or func1 == func2:
            return func1

        func_diff = func1.diff(func2)
        merge_func = func1.copy()

        if merge_func.header is None:
            merge_func.header = func2.header.copy() if func2.header else None
        elif func2.header is not None:
            merge_func.header = merge_func.header.nonconflict_merge(func2.header)

        # stack vars
        for off, var in func2.stack_vars.items():
            merge_var = func1.stack_vars[off].copy() if off in func1.stack_vars else var
            merge_var = StackVariable.nonconflict_merge(merge_var, var)

            merge_func.stack_vars[off] = merge_var

        return merge_func

    #
    # Property Shortcuts (Alias)
    #

    @property
    def name(self):
        return self.header.name if self.header else None

    @name.setter
    def name(self, value):
        # create a header if one does not exist for this function
        if not self.header:
            self.header = FunctionHeader(None, self.addr)
        self.header.name = value

    @property
    def args(self):
        return self.header.args

    def set_stack_var(self, name, off: int, off_type: int, size: int, type_str, last_change):
        self.stack_vars[off] = StackVariable(off, off_type, name, type_str, size, self.addr, last_change=last_change)



# ----------------------------------------------------------------------------
# This file is more of a library for making compatibility calls to IDA for
# things such as getting decompiled function names, start addresses, and
# asking for write permission to ida. This will mostly be called in the
# controller.
#
# Note that anything that requires write permission to IDA will need to pass
# through this program if it is not running in the main thread.
#
# ----------------------------------------------------------------------------

import threading
from functools import wraps
import typing
import logging
from time import time, sleep

import idc, idaapi, ida_kernwin, ida_hexrays, ida_funcs, \
    ida_bytes, ida_struct, ida_idaapi, ida_typeinf, idautils, ida_enum

import binsync
from binsync.data import (
    Struct, FunctionHeader, FunctionArgument, StackVariable, StackOffsetType, Function, GlobalVariable, Enum
)
from .controller import IDABinSyncController

l = logging.getLogger(__name__)

#
# Wrappers for IDA Main thread r/w operations
# a special note about these functions:
# Any operation that needs to do some type of write to the ida db (idb), needs to be in the main thread due to
# some ida constraints. Sometimes reads also need to be in the main thread. To make things efficient, most heavy
# things are done in the controller and just setters and getters are done here.
#


def is_mainthread():
    """
    Return a bool that indicates if this is the main application thread.
    """
    return isinstance(threading.current_thread(), threading._MainThread)


def execute_write(f):
    @wraps(f)
    def _execute_write(*args, **kwargs):
        output = [None]

        def thunk():
            output[0] = f(*args, **kwargs)
            return 1

        if is_mainthread():
            thunk()
        else:
            idaapi.execute_sync(thunk, idaapi.MFF_WRITE)

        # return the output of the synchronized execution
        return output[0]

    return _execute_write

#
#   Data Type Converters
#

@execute_write
def convert_type_str_to_ida_type(type_str) -> typing.Optional['ida_typeinf']:
    ida_type_str = type_str + ";"
    tif = ida_typeinf.tinfo_t()
    valid_parse = ida_typeinf.parse_decl(tif, None, ida_type_str, 1)

    return tif if valid_parse is not None else None


@execute_write
def ida_to_angr_stack_offset(func_addr, ida_stack_off):
    frame = idaapi.get_frame(func_addr)
    if not frame:
        return ida_stack_off

    frame_size = idc.get_struc_size(frame)
    last_member_size = idaapi.get_member_size(frame.get_member(frame.memqty - 1))
    angr_stack_off = ida_stack_off - frame_size + last_member_size
    return angr_stack_off


@execute_write
def convert_size_to_flag(size):
    """
    Converts a size to the arch specific flag.

    Inspired by: https://github.com/arizvisa/ida-minsc/blob/master/base/_interface.py

    :param size: in bytes
    :return: ida flag_t
    """

    size_map = {
        1: idaapi.byte_flag(),
        2: idaapi.word_flag(),
        4: idaapi.dword_flag(),
        8: idaapi.qword_flag()
    }

    try:
        flag = size_map[size]
    except KeyError:
        # just always assign something
        flag = idaapi.byte_flag()

    return flag


#
#   IDA Function r/w
#

@execute_write
def ida_func_addr(addr):
    ida_func = ida_funcs.get_func(addr)
    if ida_func is None:
        return None

    func_addr = ida_func.start_ea
    return func_addr


@execute_write
def get_func_name(ea) -> typing.Optional[str]:
    return idc.get_func_name(ea)


@execute_write
def get_func_size(ea):
    func = idaapi.get_func(ea)
    if not func:
        return 0

    return func.size()


@execute_write
def set_ida_func_name(func_addr, new_name):
    idaapi.set_name(func_addr, new_name, idaapi.SN_FORCE)
    ida_kernwin.request_refresh(ida_kernwin.IWID_DISASMS)
    ida_kernwin.request_refresh(ida_kernwin.IWID_STRUCTS)
    ida_kernwin.request_refresh(ida_kernwin.IWID_STKVIEW)


@execute_write
def functions():
    blacklisted_segs = ["extern", ".plt", ".plt.sec"]
    func_addrs = list(idautils.Functions())
    funcs = {}
    for func_addr in func_addrs:
        # skip non-text segments
        if idc.get_segm_name(func_addr) in blacklisted_segs:
            continue

        func_name = get_func_name(func_addr)
        func_size = get_func_size(func_addr)
        func = Function(func_addr, func_size)
        func.name = func_name
        funcs[func_addr] = func

    return funcs


@execute_write
def function(addr, decompiler_available=True):
    ida_func = ida_funcs.get_func(addr)
    if ida_func is None:
        l.warning(f"IDA function does not exist for {hex(addr)}.")
        return None

    func_addr = ida_func.start_ea

    if decompiler_available:
        try:
            ida_cfunc = idaapi.decompile(func_addr)
        except Exception:
            ida_cfunc = None
    else:
        ida_cfunc = None

    change_time = int(time())
    func = Function(func_addr, get_func_size(func_addr), last_change=change_time)
    if ida_cfunc is None:
        if decompiler_available:
            l.warning(f"IDA function {hex(func_addr)} is not decompilable")

        func.header = FunctionHeader(get_func_name(func_addr), func_addr, last_change=change_time)
        return func

    # decompilation is available
    func_header: FunctionHeader = function_header(ida_cfunc)
    stack_vars = {
        offset: var
        for offset, var in get_func_stack_var_info(ida_func.start_ea).items()
    }
    func.header = func_header
    func.stack_vars = stack_vars

    return func


@execute_write
def function_header(ida_cfunc) -> FunctionHeader:
    func_addr = ida_cfunc.entry_ea

    # collect the function arguments
    func_args = {}
    for idx, arg in enumerate(ida_cfunc.arguments):
        size = arg.width
        name = arg.name
        type_str = str(arg.type())
        func_args[idx] = FunctionArgument(idx, name, type_str, size)

    # collect the header ret_type and name
    func_name = get_func_name(func_addr)
    try:
        ret_type_str = str(ida_cfunc.type.get_rettype())
    except Exception:
        ret_type_str = ""

    ida_function_info = FunctionHeader(func_name, func_addr, ret_type=ret_type_str, args=func_args, last_change=int(time()))
    return ida_function_info


@execute_write
def set_function_header(ida_func_code_view, binsync_header: binsync.data.FunctionHeader, exit_on_bad_type=False):
    data_changed = False
    ida_cfunc = ida_func_code_view.cfunc
    func_addr = ida_cfunc.entry_ea

    cur_ida_func = function_header(ida_cfunc)

    #
    # FUNCTION NAME
    #

    if binsync_header.name and binsync_header.name != cur_ida_func.name:
        set_ida_func_name(func_addr, binsync_header.name)

    #
    # FUNCTION RET TYPE
    #

    func_name = get_func_name(func_addr)
    cur_ret_type_str = str(ida_cfunc.type.get_rettype())
    if binsync_header.ret_type and binsync_header.ret_type != cur_ret_type_str:
        old_prototype = str(ida_cfunc.type).replace("(", f" {func_name}(", 1)
        new_prototype = old_prototype.replace(cur_ret_type_str, binsync_header.ret_type, 1)
        success = idc.SetType(func_addr, new_prototype)

        # we may need to reload types
        if success is None and exit_on_bad_type:
            return None

        data_changed |= success is True
        refresh_pseudocode_view(func_addr)

    #
    # FUNCTION ARGS
    #

    types_to_change = {}
    for idx, binsync_arg in binsync_header.args.items():
        if idx >= len(cur_ida_func.args):
            break

        cur_ida_arg = cur_ida_func.args[idx]

        # change the name
        if binsync_arg.name and binsync_arg.name != cur_ida_arg.name:
            success = ida_func_code_view.rename_lvar(ida_cfunc.arguments[idx], binsync_arg.name, 1)
            data_changed |= success is True
            refresh_pseudocode_view(func_addr)

        # record the type to change
        if binsync_arg.type_str and binsync_arg.type_str != cur_ida_arg.type_str:
            types_to_change[idx] = (cur_ida_arg.type_str, binsync_arg.type_str)

    # crazy prototype parsing
    func_prototype = str(ida_cfunc.type).replace("(", f" {func_name}(", 1)
    proto_split = func_prototype.split("(", maxsplit=1)
    proto_head, proto_body = proto_split[0], "(" + proto_split[1]
    arg_strs = proto_body.split(",")

    # update prototype body from left to right
    for idx in range(len(cur_ida_func.args)):
        try:
            old_t, new_t = types_to_change[idx]
        except KeyError:
            continue

        arg_strs[idx] = arg_strs[idx].replace(old_t, new_t, 1)

    # set the change
    proto_body = ",".join(arg_strs)
    new_prototype = proto_head + proto_body
    success = idc.SetType(func_addr, new_prototype)

    # we may need to reload types
    if success is None and exit_on_bad_type:
        return None

    data_changed |= success is True

    return data_changed


#
#   IDA Comment r/w
#

@execute_write
def set_ida_comment(addr, cmt, decompiled=False):
    func = ida_funcs.get_func(addr)
    if not func:
        l.info(f"No function found at {addr}")
        return False

    rpt = 1

    # function comment
    if addr == func.start_ea:
        idc.set_func_cmt(addr, cmt, rpt)
        return True

    # a comment in decompilation
    elif decompiled:
        try:
            cfunc = idaapi.decompile(addr)
        except Exception:
            ida_bytes.set_cmt(addr, cmt, rpt)
            return True

        eamap = cfunc.get_eamap()
        decomp_obj_addr = eamap[addr][0].ea
        tl = idaapi.treeloc_t()

        # try to set a comment using the cfunc obj and normal address
        for a in [addr, decomp_obj_addr]:
            tl.ea = a
            for itp in range(idaapi.ITP_SEMI, idaapi.ITP_COLON):
                tl.itp = itp
                cfunc.set_user_cmt(tl, cmt)
                cfunc.save_user_cmts()
                cfunc.refresh_func_ctext()

                # attempt to set until it does not fail (orphan itself)
                if not cfunc.has_orphan_cmts():
                    cfunc.save_user_cmts()
                    return True
                cfunc.del_orphan_cmts()

        return False

    # a comment in disassembly
    else:
        ida_bytes.set_cmt(addr, cmt, rpt)
        return True


@execute_write
def set_decomp_comments(func_addr, cmt_dict: typing.Dict[int, str]):
    for addr in cmt_dict:
        ida_cmts = ida_hexrays.user_cmts_new()

        comment = cmt_dict[addr]
        tl = ida_hexrays.treeloc_t()
        tl.ea = addr
        # XXX: need a real value here at some point
        tl.itp = 90
        ida_cmts.insert(tl, ida_hexrays.citem_cmt_t(comment))

        ida_hexrays.save_user_cmts(func_addr, ida_cmts)


#
#   IDA Stack Var r/w
#

@execute_write
def get_func_stack_var_info(func_addr) -> typing.Dict[int, StackVariable]:
    try:
        decompilation = ida_hexrays.decompile(func_addr)
    except ida_hexrays.DecompilationFailure:
        l.debug("Decompiling too many functions too fast! Slow down and try that operation again.")
        return {}

    stack_var_info = {}

    for var in decompilation.lvars:
        if not var.is_stk_var():
            continue

        size = var.width
        name = var.name
        
        ida_offset = var.location.stkoff() - decompilation.get_stkoff_delta()
        bs_offset = ida_to_angr_stack_offset(func_addr, ida_offset)
        type_str = str(var.type())
        stack_var_info[bs_offset] = StackVariable(
            ida_offset, StackOffsetType.IDA, name, type_str, size, func_addr
        )

    return stack_var_info


@execute_write
def set_stack_vars_types(var_type_dict, code_view, controller: "IDABinSyncController") -> bool:
    """
    Sets the type of a stack variable, which should be a local variable.
    Take special note of the types of first two parameters used here:
    var_type_dict is a dictionary of the offsets and the new proposed type info for each offset.
    This typeinfo should be gotten either by manully making a new typeinfo object or using the
    parse_decl function. code_view is a _instance of vdui_t, which should be gotten through
    open_pseudocode() from ida_hexrays.

    This function also is special since it needs to iterate all of the stack variables an unknown amount
    of times until a fixed point of variables types not changing is met.


    @param var_type_dict:       Dict[stack_offset, ida_typeinf_t]
    @param code_view:           A pointer to a vdui_t screen
    @param controller:          The BinSync controller to do operations on
    @return:
    """

    data_changed = False
    fixed_point = False
    while not fixed_point:
        fixed_point = True
        for lvar in code_view.cfunc.lvars:
            cur_off = lvar.location.stkoff() - code_view.cfunc.get_stkoff_delta()
            if lvar.is_stk_var() and cur_off in var_type_dict:
                if str(lvar.type()) != str(var_type_dict[cur_off]):
                    data_changed |= code_view.set_lvar_type(lvar, var_type_dict.pop(cur_off))
                    fixed_point = False
                    # make sure to break, in case the size of lvars array has now changed
                    break

    return data_changed

@execute_write
def ida_get_frame(func_addr):
    return idaapi.get_frame(func_addr)


#
#   IDA Struct r/w
#

@execute_write
def structs():
    _structs = {}
    for struct_item in idautils.Structs():
        idx, sid, name = struct_item[:]
        sptr = ida_struct.get_struc(sid)
        size = ida_struct.get_struc_size(sptr)
        _structs[name] = Struct(name, size, {})
        
    return _structs

@execute_write
def struct(name):
    sid = ida_struct.get_struc_id(name)
    if sid == 0xffffffffffffffff:
        return None
    
    sptr = ida_struct.get_struc(sid)
    size = ida_struct.get_struc_size(sptr)
    _struct = Struct(name, size, {}, last_change=int(time()))
    for mptr in sptr.members:
        mid = mptr.id
        m_name = ida_struct.get_member_name(mid)
        m_off = mptr.soff
        m_type = ida_typeinf.idc_get_type(mptr.id) if mptr.has_ti() else ""
        m_size = ida_struct.get_member_size(mptr)
        _struct.add_struct_member(m_name, m_off, m_type, m_size)

    return _struct

@execute_write
def set_struct_member_name(ida_struct, frame, offset, name):
    ida_struct.set_member_name(frame, offset, name)

@execute_write
def set_ida_struct(struct: Struct, controller) -> bool:
    # first, delete any struct by the same name if it exists
    sid = ida_struct.get_struc_id(struct.name)
    if sid != 0xffffffffffffffff:
        sptr = ida_struct.get_struc(sid)
        ida_struct.del_struc(sptr)

    # now make a struct header
    ida_struct.add_struc(ida_idaapi.BADADDR, struct.name, False)
    sid = ida_struct.get_struc_id(struct.name)
    sptr = ida_struct.get_struc(sid)

    # expand the struct to the desired size
    # XXX: do not increment API here, why? Not sure, but you cant do it here.
    ida_struct.expand_struc(sptr, 0, struct.size)

    # add every member of the struct
    for off, member in struct.struct_members.items():
        # convert to ida's flag system
        mflag = convert_size_to_flag(member.size)

        # create the new member
        ida_struct.add_struc_member(
            sptr,
            member.member_name,
            member.offset,
            mflag,
            None,
            member.size,
        )

    return True


@execute_write
def set_ida_struct_member_types(struct: Struct, controller) -> bool:
    # find the specific struct
    sid = ida_struct.get_struc_id(struct.name)
    sptr = ida_struct.get_struc(sid)
    data_changed = False

    for idx, member in enumerate(struct.struct_members.values()):
        # set the new member type if it has one
        if member.type == "":
            continue

        # assure its convertible
        tif = convert_type_str_to_ida_type(member.type)
        if tif is None:
            continue

        # set the type
        mptr = sptr.get_member(idx)
        was_set = ida_struct.set_member_tinfo(
            sptr,
            mptr,
            0,
            tif,
            mptr.flag
        )
        data_changed |= was_set == 1

    return data_changed

#
# Global Vars
#


@execute_write
def global_vars():
    gvars = {}
    known_segs = [".data", ".bss"]
    for seg_name in known_segs:
        seg = idaapi.get_segm_by_name(seg_name)
        if not seg:
            continue

        for seg_ea in range(seg.start_ea, seg.end_ea):
            xrefs = idautils.XrefsTo(seg_ea)
            try:
                next(xrefs)
            except StopIteration:
                continue

            name = idaapi.get_name(seg_ea)
            if not name:
                continue

            gvars[seg_ea] = GlobalVariable(seg_ea, name)

    return gvars


@execute_write
def global_var(addr):
    name = idaapi.get_name(addr)
    if not name:
        return None

    size = idaapi.get_item_size(addr)
    return GlobalVariable(addr, name, size=size, last_change=int(time()))


@execute_write
def set_global_var_name(var_addr, name):
    return idaapi.set_name(var_addr, name)

#
# Enums
#

def get_enum_members(_enum) -> typing.Dict[str, int]:
    enum_members = {}

    member = ida_enum.get_first_enum_member(_enum)
    member_addr = ida_enum.get_enum_member(_enum, member, 0, 0)
    member_name = ida_enum.get_enum_member_name(member_addr)
    enum_members[member_name] = member

    while member := ida_enum.get_next_enum_member(_enum, member, 0):
        if member == 0xffffffffffffffff: break
        member_addr = ida_enum.get_enum_member(_enum, member, 0, 0)
        member_name = ida_enum.get_enum_member_name(member_addr)
        enum_members[member_name] = member
    return enum_members


@execute_write
def enums() -> typing.Dict[str, Enum]:
    _enums: typing.Dict[str, Enum] = {}
    for i in range(ida_enum.get_enum_qty()):
        _enum = ida_enum.getn_enum(i)
        enum_name = ida_enum.get_enum_name(_enum)
        enum_members = get_enum_members(_enum)
        _enums[enum_name] = Enum(enum_name, enum_members)
    return _enums


@execute_write
def enum(name) -> typing.Optional[Enum]:
    _enum = ida_enum.get_enum(name)
    if not _enum:
        return None
    enum_name = ida_enum.get_enum_name(_enum)
    enum_members = get_enum_members(_enum)
    return Enum(enum_name, enum_members)


@execute_write
def set_enum(bs_enum: Enum):
    _enum = ida_enum.get_enum(bs_enum.name)
    if not _enum:
        return False

    ida_enum.del_enum(_enum)
    enum_id = ida_enum.add_enum(ida_enum.get_enum_qty(), bs_enum.name, 0)

    if enum_id is None:
        l.warning(f"IDA failed to create a new enum with {bs_enum.name}")
        return False

    for member_name, value in bs_enum.members.items():
        ida_enum.add_enum_member(enum_id, member_name, value)

    return True


#
#   IDA GUI r/w
#

@execute_write
def acquire_pseudocode_vdui(addr):
    """
    Acquires a IDA HexRays vdui pointer, which is a pointer to a pseudocode view that contains
    the cfunc which describes the code on the screen. Using this function optimizes the switching of code views
    by using in-place switching if a view is already present.

    @param addr:
    @return:
    """
    func = ida_funcs.get_func(addr)
    if not func:
        return None

    names = ["Pseudocode-%c" % chr(ord("A") + i) for i in range(5)]
    for name in names:
        widget = ida_kernwin.find_widget(name)
        if not widget:
            continue

        vu = ida_hexrays.get_widget_vdui(widget)
        break
    else:
        vu = ida_hexrays.open_pseudocode(func.start_ea, False)

    if func.start_ea != vu.cfunc.entry_ea:
        target_cfunc = idaapi.decompile(func.start_ea)
        vu.switch_to(target_cfunc, False)

    return vu


@execute_write
def refresh_pseudocode_view(ea, set_focus=True):
    """Refreshes the pseudocode view in IDA."""
    names = ["Pseudocode-%c" % chr(ord("A") + i) for i in range(5)]
    for name in names:
        widget = ida_kernwin.find_widget(name)
        if widget:
            vu = ida_hexrays.get_widget_vdui(widget)

            # Check if the address is in the same function
            func_ea = vu.cfunc.entry_ea
            func = ida_funcs.get_func(func_ea)
            if ida_funcs.func_contains(func, ea):
                vu.refresh_view(True)
                ida_kernwin.activate_widget(widget, set_focus)


class IDAViewCTX:
    @execute_write
    def __init__(self, func_addr):
        self.view = ida_hexrays.open_pseudocode(func_addr, 0)

    def __enter__(self):
        return self.view

    @execute_write
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_pseudocode_view(self.view)

    @execute_write
    def close_pseudocode_view(self, ida_vdui_t):
        widget = ida_vdui_t.toplevel
        idaapi.close_pseudocode(widget)


def get_screen_ea():
    return idc.get_screen_ea()


@execute_write
def get_function_cursor_at():
    curr_addr = get_screen_ea()
    if curr_addr is None:
        return None

    return ida_func_addr(curr_addr)


#
# Other Utils
#

@execute_write
def get_ptr_size():
    """
    Gets the size of the ptr, which in affect tells you the bit size of the binary.

    Taken from: https://github.com/arizvisa/ida-minsc/blob/master/base/database.py
    :return: int, size in bytes
    """
    tif = ida_typeinf.tinfo_t()
    tif.create_ptr(ida_typeinf.tinfo_t(ida_typeinf.BT_VOID))
    return tif.get_size()


@execute_write
def get_binary_path():
    return idaapi.get_input_file_path()


@execute_write
def jumpto(addr):
    """
    Changes the pseudocode view to the function address provided.

    @param addr: Address of function to jump to
    @return:
    """
    idaapi.jumpto(addr)



import uuid


class User:
    """
    :ivar str name: Name of the user
    :ivar str uid:  Internal user ID in the form of uuid.
    :ivar int push_time: Last push time of user.
    :ivar int last_push_func: Last function address modified pushed.
    """

    def __init__(self, name, uid=None, client=None,
                 last_push_time=-1, last_push_artifact=-1, last_push_artifact_type=1):
        self.name = name
        self.uid = uid if uid is not None else uuid.uuid4()
        self.client = client
        self.last_push_time = last_push_time
        self.last_push_artifact = last_push_artifact
        self.last_push_artifact_type = last_push_artifact_type

    @classmethod
    def from_metadata(cls, metadata):
        u = cls(
            metadata["user"],
            uid=metadata.get("uid", None),
            client=metadata.get("client", None),
            last_push_time=metadata.get("last_push_time", -1),
            last_push_artifact=metadata.get("last_push_artifact", -1),
            last_push_artifact_type=metadata.get("last_push_artifact_type", -1)
        )
        return u

    def copy(self):
        return User(
            self.name,
            uid=self.uid,
            client=self.client,
            last_push_time=self.last_push_time,
            last_push_artifact=self.last_push_artifact,
            last_push_artifact_type=self.last_push_artifact_type
        )



# ----------------------------------------------------------------------------
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
# This program describes each hook in IDA that we want to overwrite on the
# startup of IDA. Each hook function/class describes a different scenario
# that we try to track when a user makes a change. For _instance, the function
# `cmt_changed` is activated every time a user changes a disassembly comment,
# allowing us to send the new comment to be queued in the Controller actions.
#
# ----------------------------------------------------------------------------
import threading
import time
from functools import wraps
import logging

import ida_auto
import ida_bytes
import ida_enum
import ida_funcs
import ida_hexrays
import ida_idaapi
import ida_idp
import ida_kernwin
import ida_nalt
import ida_netnode
import ida_pro
import ida_segment
import ida_struct
import ida_typeinf
import ida_enum
import idaapi
import idc

from . import compat
from .controller import IDABinSyncController
from binsync.data import (
    Function, FunctionHeader, FunctionArgument, StackVariable, StackOffsetType,
    Comment, GlobalVariable, Patch,
    Enum, Struct
)

l = logging.getLogger(__name__)

#
#   Decorators
#


def quite_init_checker(f):
    @wraps(f)
    def initcheck(self, *args, **kwargs):
        if not self.controller.check_client():
            return 0
        return f(self, *args, **kwargs)
    return initcheck


def stop_if_syncing(f):
    @wraps(f)
    def _stop_if_syncing(self, *args, **kwargs):
        if self.controller.sync_lock.locked():
            return 0

        return f(self, *args, **kwargs)

    return _stop_if_syncing

#
#   IDA Change Hooks
#


class IDBHooks(ida_idp.IDB_Hooks):
    def __init__(self, controller):
        ida_idp.IDB_Hooks.__init__(self)
        self.controller: IDABinSyncController = controller
        self.last_local_type = None

    @quite_init_checker
    @stop_if_syncing
    def local_types_changed(self):
        #print("local type changed")
        return 0

    @quite_init_checker
    @stop_if_syncing
    def ti_changed(self, ea, type_, fname):
        #print(f"TI CHANGED: {ea}, {type_}, {fname}")
        return 0

    #
    #   Enum Hooks
    #

    def bs_enum_modified(self, enum):
        name = ida_enum.get_enum_name(enum)
        _enum = compat.enum(name)
        self.binsync_state_change(
            self.controller.push_artifact,
            _enum
        )

    @quite_init_checker
    @stop_if_syncing
    def enum_created(self, enum):
        self.bs_enum_modified(enum)
        return 0

    # XXX - use enum_deleted(self, id) instead?
    @quite_init_checker
    @stop_if_syncing
    def deleting_enum(self, id):
        name = ida_enum.get_enum_name(id)
        enum = Enum(name, {})
        self.binsync_state_change(
            self.controller.push_artifact,
            enum
        )
        return 0

    # XXX - use enum_renamed(self, id) instead?
    @quite_init_checker
    @stop_if_syncing
    def renaming_enum(self, id, is_enum, newname):
        if is_enum:
            self.bs_enum_modified(id)
        else:
            self.bs_enum_modified(ida_enum.get_enum_member_enum(id))
        return 0

    @quite_init_checker
    @stop_if_syncing
    def enum_bf_changed(self, id):
        #print("enum renamed")
        return 0

    @quite_init_checker
    @stop_if_syncing
    def enum_cmt_changed(self, tid, repeatable_cmt):
        #print("enum renamed")
        return 0

    @quite_init_checker
    @stop_if_syncing
    def enum_member_created(self, id, cid):
        self.bs_enum_modified(id)
        return 0

    # XXX - use enum_member_deleted(self, id, cid) instead?
    @quite_init_checker
    @stop_if_syncing
    def deleting_enum_member(self, id, cid):
        self.bs_enum_modified(id)
        return 0

    #
    #   Struct Hooks
    #

    @quite_init_checker
    @stop_if_syncing
    def struc_created(self, tid):
        #print("struct created")
        sptr = ida_struct.get_struc(tid)
        if not sptr.is_frame():
            self.ida_struct_changed(tid, old_name="")
        return 0

    # XXX - use struc_deleted(self, struc_id) instead?
    @quite_init_checker
    @stop_if_syncing
    def deleting_struc(self, sptr):
        if not sptr.is_frame():
            self.ida_struct_changed(sptr.id, deleted=True)
        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_align_changed(self, sptr):
        if not sptr.is_frame():
            self.ida_struct_changed(sptr.id)

        return 0

    # XXX - use struc_renamed(self, sptr) instead?
    @quite_init_checker
    @stop_if_syncing
    def renaming_struc(self, id, oldname, newname):
        sptr = ida_struct.get_struc(id)
        if not sptr.is_frame():
            self.ida_struct_changed(id, old_name=oldname, new_name=newname)
        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_expanded(self, sptr):
        #print("struct expanded")
        if not sptr.is_frame():
            self.ida_struct_changed(sptr.id)

        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_member_created(self, sptr, mptr):
        #print("struc member created")
        if not sptr.is_frame():
            self.ida_struct_changed(sptr.id)

        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_member_deleted(self, sptr, off1, off2):
        #print("struc member deleted")
        if not sptr.is_frame():
            self.ida_struct_changed(sptr.id)

        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_member_renamed(self, sptr, mptr):
        #print(f"struc member renamed: {sptr.id}: {mptr.id}")
        """
        Handles renaming of two things:
        1. Global Structs
        2. Stack Variables

        :param sptr:    Struct Pointer
        :param mptr:    Member Pointer
        :return:
        """
        # struct pointer is actually a stack frame
        if sptr.is_frame():
            stack_frame = sptr
            func_addr = idaapi.get_func_by_frame(stack_frame.id)
            try:
                stack_var_info = compat.get_func_stack_var_info(func_addr)[compat.ida_to_angr_stack_offset(func_addr, mptr.soff)]
            except KeyError:
                l.debug(f"Failed to track an internal changing stack var: {mptr.id}.")
                return 0

            # find the properties of the changed stack var
            angr_offset = compat.ida_to_angr_stack_offset(func_addr, stack_var_info.stack_offset)
            size = stack_var_info.size
            type_str = stack_var_info.type

            # TODO: correct this fix in the get_func_stack_var_info
            new_name = ida_struct.get_member_name(mptr.id)

            # do the change on a new thread
            sv = StackVariable(
                angr_offset, StackOffsetType.IDA, new_name, type_str, size, func_addr
            )
            self.binsync_state_change(
                self.controller.push_artifact,
                sv
            )

        # an actual struct
        else:
            self.ida_struct_changed(sptr.id)

        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_member_changed(self, sptr, mptr):
        #print(f"struc member changed: {sptr.id}, {mptr.id}")

        # struct pointer is actually a stack frame
        if sptr.is_frame():
            stack_frame = sptr
            func_addr = idaapi.get_func_by_frame(stack_frame.id)
            try:
                all_var_info = compat.get_func_stack_var_info(func_addr)
                stack_var_info = all_var_info[compat.ida_to_angr_stack_offset(func_addr, mptr.soff)]
            except KeyError:
                l.debug(f"Failed to track an internal changing stack var: {mptr.id}.")
                return 0

            # find the properties of the changed stack var
            angr_offset = compat.ida_to_angr_stack_offset(func_addr, stack_var_info.stack_offset)
            size = stack_var_info.size
            type_str = stack_var_info.type

            new_name = stack_var_info.name #ida_struct.get_member_name(mptr.id)

            # do the change on a new thread
            sv = StackVariable(
                angr_offset, StackOffsetType.IDA, new_name, type_str, size, func_addr
            )
            self.binsync_state_change(
                self.controller.push_artifact,
                sv
            )
        else:
            self.ida_struct_changed(sptr.id)

        return 0

    @quite_init_checker
    @stop_if_syncing
    def struc_cmt_changed(self, id, repeatable_cmt):
        fullname = ida_struct.get_struc_name(id)
        if "." in fullname:
            sname, smname = fullname.split(".", 1)
        else:
            sname = fullname
            smname = ""
        cmt = ida_struct.get_struc_cmt(id, repeatable_cmt)
        return 0

    @quite_init_checker
    @stop_if_syncing
    def sgr_changed(self, start_ea, end_ea, regnum, value, old_value, tag):
        # FIXME: sgr_changed is not triggered when a segment register is
        # being deleted by the user, so we need to sent the complete list
        return 0

    @quite_init_checker
    @stop_if_syncing
    def renamed(self, ea, new_name, local_name):
        # #print("renamed(ea = %x, new_name = %s, local_name = %d)" % (ea, new_name, local_name))
        if ida_struct.is_member_id(ea) or ida_struct.get_struc(ea) or ida_enum.get_enum_name(ea):
            return 0

        ida_func = idaapi.get_func(ea)
        # global var renaming
        if ida_func is None:
            size = idaapi.get_item_size(ea)
            self.binsync_state_change(
                self.controller.push_artifact,
                GlobalVariable(ea, new_name, size=size)
            )

        # function name renaming
        elif ida_func.start_ea == ea:
            # grab the name instead from ida
            name = idc.get_func_name(ida_func.start_ea)
            self.binsync_state_change(
                self.controller.push_artifact,
                FunctionHeader(name, ida_func.start_ea)
            )

        return 0

    @quite_init_checker
    @stop_if_syncing
    def byte_patched(self, ea, old_value):
        return 0

    @quite_init_checker
    @stop_if_syncing
    def cmt_changed(self, ea, repeatable_cmt):
        if repeatable_cmt:
            cmt = ida_bytes.get_cmt(ea, repeatable_cmt)
            if cmt:
                self.ida_comment_changed(cmt, ea, "cmt")
        return 0

    @quite_init_checker
    @stop_if_syncing
    def range_cmt_changed(self, kind, a, cmt, repeatable):
        #print("range cmt changed")
        # verify it's a function comment
        cmt = idc.get_func_cmt(a.start_ea, repeatable)
        if cmt:
            self.ida_comment_changed(cmt, a.start_ea, "range")

        return 0

    @quite_init_checker
    @stop_if_syncing
    def extra_cmt_changed(self, ea, line_idx, cmt):
        #print("extra cmt changed")
        cmt = ida_bytes.get_cmt(ea, 0)
        if cmt:
            self.ida_comment_changed(cmt, ea, "cmt")
        return 0

    #
    #   Helpers
    #

    def ida_comment_changed(self, comment: str, address: int, cmt_type: str):
        """
        Utility function to catch all types of comment changes.

        @param comment:
        @param address:
        @param cmt_type:
        @return:
        """
        ida_func = idaapi.get_func(address)
        func_addr = ida_func.start_ea if ida_func else None
        kwarg = {"func_addr": func_addr}

        bs_cmt = Comment(address, comment, **kwarg)
        # disass comment changed
        if cmt_type == "cmt":
            self.binsync_state_change(
                self.controller.push_artifact,
                bs_cmt
            )

        # function comment changed
        elif cmt_type == "range":
            # overwrite the entire function comment
            self.binsync_state_change(
                self.controller.push_artifact,
                bs_cmt
            )

        # XXX: other?
        elif cmt_type == "extra":
            return 0

        return 0

    def ida_struct_changed(self, sid: int, old_name=None, new_name=None, deleted=False):
        """
        A utility function to catch all changes that can happen to a struct:
        1. Renames
        2. Member Changes
        3. Deletes

        Currently, any change to a struct will cause the main-thread to re-copy the entire struct from the local
        state into the remote state. This is done so we don't need to have multiple cases for single member changes.

        @param sid:         Struct ID (IDA Thing)
        @param old_name:    Old struct name (before rename)
        @param new_name:    New struct name (after rename)
        @param deleted:     True only when the entire struct has been deleted.
        @return:
        """
        # parse the info of the current struct
        s_name = new_name if new_name else ida_struct.get_struc_name(sid)

        # back out if a stack variable snuck in
        if s_name.startswith("$"):
            return 0

        sptr = ida_struct.get_struc(sid)
        s_size = ida_struct.get_struc_size(sptr)

        # if deleted, finish early
        if deleted:
            self.binsync_state_change(
                self.controller.push_artifact,
                Struct(s_name, None, {})
            )
            return 0

        # convert the ida_struct into a binsync_struct
        binsync_struct = Struct(s_name, s_size, {})
        for mptr in sptr.members:
            mid = mptr.id
            m_name = ida_struct.get_member_name(mid)
            m_off = mptr.soff
            m_type = ida_typeinf.idc_get_type(mptr.id) if mptr.has_ti() else ""
            m_size = ida_struct.get_member_size(mptr)
            binsync_struct.add_struct_member(m_name, m_off, m_type, m_size)

        # make the controller update the local state and push
        old_s_name = old_name if old_name else s_name
        self.binsync_state_change(
            self.controller.push_artifact,
            binsync_struct
        )
        return 0

    def binsync_state_change(self, *args, **kwargs):
        self.controller.schedule_job(*args, **kwargs)


class IDPHooks(ida_idp.IDP_Hooks):
    def __init__(self, controller):
        self.controller = controller
        ida_idp.IDP_Hooks.__init__(self)

    def ev_adjust_argloc(self, *args):
        return ida_idp.IDP_Hooks.ev_adjust_argloc(self, *args)

    def ev_ending_undo(self, action_name, is_undo):
        """
        This is the hook called by IDA when an undo event occurs
        action name is a vague String description of what changes occured
        is_undo specifies if this action was an undo or a redo
        """
        return 0

    def ev_replaying_undo(self, action_name, vec, is_undo):
        """
        This hook is also called by IDA during the undo
        contains the same information as ev_ending_undo
        vec also contains a short summary of changes incurred
        """
        return 0

class HexRaysHooks:
    def __init__(self, controller):
        self.controller: IDABinSyncController = controller
        super(HexRaysHooks, self).__init__()
        self._available = None
        self._installed = False
        self._cached_funcs = {}
        self.updating_states = threading.Lock()

    def hook(self):
        if not self.controller.decompiler_available:
            return

        ida_hexrays.install_hexrays_callback(self._hxe_callback)
        self._available = True
        self._installed = True

    def unhook(self):
        if self._available:
            self._installed = False

    @quite_init_checker
    @stop_if_syncing
    def _hxe_callback(self, event, *args):
        if not self._installed:
            return 0

        # this event gets triggered each time that a user changes the view to
        # a different decompilation view. It will also get triggered when staying on the
        # same view but having it refreshed
        if event == ida_hexrays.hxe_func_printed:
            ida_cfunc = args[0]
            func_addr = ida_cfunc.entry_ea
            func = ida_funcs.get_func(func_addr)

            # sanity check
            if func is None:
                return 0

            # run update tasks needed for this function since we are looking at it
            if not self.updating_states.locked():
                with self.updating_states:
                    self.controller.update_states[func_addr].do_updates()

            # create a new cache for unseen funcs
            if func.start_ea not in self._cached_funcs.keys():
                self._cached_funcs[func.start_ea] = {"cmts": [], "header": None}

            # push changes viewable only in decompilation
            self._push_new_comments(func.start_ea)
            self._push_new_func_header(ida_cfunc)

        return 0

    @quite_init_checker
    def _push_new_func_header(self, ida_cfunc):
        # on first time seeing it, we dont want a push
        if not self._cached_funcs[ida_cfunc.entry_ea]["header"]:
            cur_header_str = str(ida_cfunc.type)
            self._cached_funcs[ida_cfunc.entry_ea]["header"] = cur_header_str
            return

        cur_header_str = str(ida_cfunc.type)
        if cur_header_str != self._cached_funcs[ida_cfunc.entry_ea]["header"]:
            # convert to binsync type
            cur_func_header = compat.function_header(ida_cfunc)
            binsync_args = {}
            for idx, arg in cur_func_header.args.items():
                binsync_args[idx] = FunctionArgument(idx, arg.name, arg.type_str, arg.size)

            # send the change
            self.binsync_state_change(
                self.controller.push_artifact,
                cur_func_header
            )

            self._cached_funcs[ida_cfunc.entry_ea]["header"] = cur_header_str

    @staticmethod
    def _get_user_cmts(ea):
        user_cmts = ida_hexrays.restore_user_cmts(ea)
        if user_cmts is None:
            user_cmts = ida_hexrays.user_cmts_new()
        cmts = {}
        it = ida_hexrays.user_cmts_begin(user_cmts)
        while it != ida_hexrays.user_cmts_end(user_cmts):
            tl = ida_hexrays.user_cmts_first(it)
            cmt = ida_hexrays.user_cmts_second(it)
            cmts[tl.ea] = str(cmt)

            it = ida_hexrays.user_cmts_next(it)
        ida_hexrays.user_cmts_free(user_cmts)
        return cmts

    @quite_init_checker
    def _push_new_comments(self, ea):
        # get the comments for the function
        cmts = HexRaysHooks._get_user_cmts(ea)

        # validate we dont waste time
        if len(cmts) == 0:
            return

        # never do the same push twice
        if cmts != self._cached_funcs[ea]["cmts"]:
            # thread it!
            sync_cmts = [Comment(addr, cmt, decompiled=True) for addr, cmt in cmts.items()]
            for cmt in sync_cmts:
                cmt.func_addr = ea
                self.binsync_state_change(
                    self.controller.push_artifact,
                    cmt
                )

            # cache so we don't double push a copy
            self._cached_funcs[ea]["cmts"] = cmts

    @staticmethod
    def refresh_pseudocode_view(ea):
        """Refreshes the pseudocode view in IDA."""
        names = ["Pseudocode-%c" % chr(ord("A") + i) for i in range(5)]
        for name in names:
            widget = ida_kernwin.find_widget(name)
            if widget:
                vu = ida_hexrays.get_widget_vdui(widget)

                # Check if the address is in the same function
                func_ea = vu.cfunc.entry_ea
                func = ida_funcs.get_func(func_ea)
                if ida_funcs.func_contains(func, ea):
                    vu.refresh_view(False)

    def binsync_state_change(self, *args, **kwargs):
        self.controller.schedule_job(*args, **kwargs)


class MasterHook:
    def __init__(self, controller):
        self.controller = controller

        self.idb_hook = IDBHooks(controller)
        self.idp_hook = IDPHooks(controller)
        self.hexray_hook = HexRaysHooks(controller)

    def hook(self):
        self.idb_hook.hook()
        self.idp_hook.hook()
        self.hexray_hook.hook()



# pylint: disable=wrong-import-position,wrong-import-order
import logging

import binsync
from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.workspace import Workspace
from binsync.common.ui.version import set_ui_version

set_ui_version("PySide2")
from binsync.common.ui.config_dialog import SyncConfig
from .control_panel_view import ControlPanelView
from .controller import AngrBinSyncController

from binsync.data import (
    StackVariable, StackOffsetType, Function, FunctionHeader, Comment
)

l = logging.getLogger(__name__)

class BinSyncPlugin(BasePlugin):
    """
    Controller plugin for BinSync
    """
    def __init__(self, workspace: Workspace):
        """
        The entry point for the BinSync plugin. This class is respobsible for both initializing the GUI and
        deiniting it as well. The BinSync plugin also starts the BinsyncController, which is a threaded class
        that pushes and pulls changes every so many seconds.

        @param workspace:   an AM _workspace (usually found in _instance)
        """
        super().__init__(workspace)

        # init the Sync View on load
        self.controller = AngrBinSyncController(self.workspace)
        self.control_panel_view = ControlPanelView(workspace, 'right', self.controller)

        self.controller.control_panel = self.control_panel_view

        self.sync_menu = None
        self.selected_funcs = []

    #
    # BinSync Deinit
    #

    def teardown(self):
        # destroy the sync view on deinit
        self.workspace.remove_view(self.control_panel_view)

    #
    # BinSync GUI Hooks
    #

    MENU_BUTTONS = ('Configure Binsync...', 'Toggle Binsync Panel')
    MENU_CONFIG_ID = 0
    MENU_TOGGLE_PANEL_ID = 1

    def handle_click_menu(self, idx):
        # sanity check on menu selection
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.instance.project.am_none:
            return

        mapping = {
            self.MENU_CONFIG_ID: self.open_sync_config_dialog,
            self.MENU_TOGGLE_PANEL_ID: self.toggle_sync_panel
        }

        # call option mapped to each menu pos
        mapping.get(idx)()


    def open_sync_config_dialog(self):
        if self.workspace.instance.project.am_none:
            # project does not exist yet
            return

        sync_config = SyncConfig(self.controller)
        sync_config.exec_()

        if self.controller.check_client() and self.control_panel_view not in self.workspace.view_manager.views:
            self.workspace.add_view(self.control_panel_view)

    def toggle_sync_panel(self):
        self.controller.toggle_headless()

        if self.control_panel_view.isVisible():
            self.control_panel_view.close()
            return

        self.workspace.add_view(self.control_panel_view)

    #
    #   BinSync Decompiler Hooks
    #

    # pylint: disable=unused-argument
    def handle_stack_var_renamed(self, func, offset, old_name, new_name):
        decompilation = self.controller.decompile_function(func)
        stack_var = self.controller.find_stack_var_in_codegen(decompilation, offset)
        var_type = AngrBinSyncController.stack_var_type_str(decompilation, stack_var)

        self.controller.schedule_job(
            self.controller.push_artifact,
            StackVariable(offset, StackOffsetType.ANGR, new_name, var_type, stack_var.size, func.addr)
        )
        return False

    # pylint: disable=unused-argument
    def handle_stack_var_retyped(self, func, offset, old_type, new_type):
        decompilation = self.controller.decompile_function(func)
        stack_var = self.controller.find_stack_var_in_codegen(decompilation, offset)

        self.controller.schedule_job(
            self.controller.push_artifact,
            StackVariable(offset, StackOffsetType.ANGR, stack_var.name, new_type, stack_var.size, func.addr),
        )
        return False

    # pylint: disable=unused-argument
    def handle_func_arg_renamed(self, func, offset, old_name, new_name):
        decompilation = self.controller.decompile_function(func)
        func_args = AngrBinSyncController.get_func_args(decompilation)
        func_type = decompilation.cfunc.functy.returnty.c_repr()
        bs_args = {
            i: binsync.FunctionArgument(i, var_info[0].name, var_info[1], var_info[0].size)
            for i, var_info in func_args.items()
        }

        self.controller.schedule_job(
            self.controller.push_artifact,
            FunctionHeader(func.name, func.addr, ret_type=func_type, args=bs_args)
        )
        return False

    # pylint: disable=unused-argument
    def handle_func_arg_retyped(self, func, offset, old_type, new_type):
        decompilation = self.controller.decompile_function(func)
        func_args = AngrBinSyncController.get_func_args(decompilation)
        func_type = decompilation.cfunc.functy.returnty.c_repr()
        bs_args = {
            i: binsync.FunctionArgument(i, var_info[0].name, var_info[1], var_info[0].size)
            for i, var_info in func_args.items()
        }

        self.controller.schedule_job(
            self.controller.push_artifact,
            FunctionHeader(func.name, func.addr, ret_type=func_type, args=bs_args)
        )
        return False

    # pylint: disable=unused-argument,no-self-use
    def handle_global_var_renamed(self, address, old_name, new_name):
        return False

    # pylint: disable=unused-argument,no-self-use
    def handle_global_var_retyped(self, address, old_type, new_type):
        return False

    # pylint: disable=unused-argument
    def handle_function_renamed(self, func, old_name, new_name):
        self.controller.schedule_job(
            self.controller.push_artifact,
            FunctionHeader(new_name, func.addr)
        )
        return False

    # pylint: disable=unused-argument,no-self-use
    def handle_function_retyped(self, func, old_type, new_type):
        return False

    # pylint: disable=unused-argument
    def handle_comment_changed(self, address, old_cmt, new_cmt, created: bool, decomp: bool):
        func_addr = self.controller.get_func_addr_from_addr(address)
        self.controller.schedule_job(
            self.controller.push_artifact,
            Comment(address, new_cmt, func_addr=func_addr, decompiled=decomp)
        )
        return False



import toml

from binsync.data.artifact import Artifact


class StackOffsetType:
    BINJA = 0
    IDA = 1
    GHIDRA = 2
    ANGR = 3


class StackVariable(Artifact):
    """
    Describes a stack variable for a given function.
    """

    __slots__ = (
        "last_change",
        "stack_offset",
        "stack_offset_type",
        "name",
        "type",
        "size",
        "addr",
    )

    def __init__(self, stack_offset, offset_type, name, type_, size, addr, last_change=None):
        super(StackVariable, self).__init__(last_change=last_change)
        self.stack_offset = stack_offset  # type: int
        self.stack_offset_type = offset_type  # type: int
        self.name = name  # type: str
        self.type = type_  # type: str
        self.size = size  # type: int
        self.addr = addr  # type: int
        self.last_change = last_change

    def __eq__(self, other):
        # ignore time and offset type
        if isinstance(other, StackVariable):
            return other.stack_offset == self.stack_offset \
                   and other.name == self.name \
                   and other.type == self.type \
                   and other.size == self.size \
                   and other.addr == self.addr
        return False

    def __str__(self):
        return f"<StackVar: {self.type} {self.name}; {hex(self.stack_offset)}@{hex(self.addr)}>"

    def __repr__(self):
        return self.__str__()

    def get_offset(self, offset_type):
        if offset_type == self.stack_offset_type:
            return self.stack_offset
        # conversion required
        if self.stack_offset_type in (StackOffsetType.IDA, StackOffsetType.BINJA, StackOffsetType.ANGR):
            off = self.stack_offset
        else:
            raise NotImplementedError()
        if offset_type in (StackOffsetType.IDA, StackOffsetType.BINJA):
            return off
        else:
            raise NotImplementedError()

    def copy(self):
        return StackVariable(
            self.stack_offset, self.stack_offset_type, self.name, self.type, self.size, self.addr,
            last_change=self.last_change
        )

    @classmethod
    def parse(cls, s):
        sv = StackVariable(None, None, None, None, None, None)
        sv.__setstate__(toml.loads(s))
        return sv

    @classmethod
    def load_many(cls, svs_toml):
        for sv_toml in svs_toml.values():
            sv = StackVariable(None, None, None, None, None, None)
            sv.__setstate__(sv_toml)
            yield sv

    @classmethod
    def dump_many(cls, svs):
        d = { }
        for v in sorted(svs.values(), key=lambda x: x.stack_offset):
            d["%x" % v.stack_offset] = v.__getstate__()
        return d



import codecs

import toml

from binsync.data.artifact import Artifact


class Patch(Artifact):
    """
    Describes a patch on the binary code.
    """
    __slots__ = (
        "last_change",
        "offset",
        "obj_name",
        "new_bytes",
    )

    def __init__(self, offset, new_bytes, obj_name=None, last_change=None):
        super(Patch, self).__init__(last_change=last_change)
        self.offset = offset
        self.obj_name = obj_name
        self.new_bytes = new_bytes

    def __str__(self):
        return f"<Patch: {self.obj_name}@{hex(self.offset)} len={len(self.new_bytes)}>"

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        return {
            "obj_name": self.obj_name,
            "offset": hex(self.offset),
            "new_bytes": codecs.encode(self.new_bytes, "hex"),
            "last_change": self.last_change
        }

    def __setstate__(self, state):
        self.obj_name = state["obj_name"]
        self.offset = int(state["offset"], 16)
        self.new_bytes = codecs.decode(state["new_bytes"], "hex")
        self.last_change = state.get("last_change", None)

    @classmethod
    def parse(cls, s):
        patch = Patch(None, None, None)
        patch.__setstate__(toml.loads(s))
        return patch

    @classmethod
    def load_many(cls, patches_toml):
        for patch_toml in patches_toml.values():
            patch = Patch(None, None, None)
            try:
                patch.__setstate__(patch_toml)
            except TypeError:
                # skip all incorrect ones
                continue
            yield patch

    @classmethod
    def dump_many(cls, patches):
        patches_ = {}
        for v in patches.values():
            patches_["%s_%x" % (v.obj_name, v.offset)] = v.__getstate__()
        return patches_

    def copy(self):
        return Patch(
            self.offset,
            self.new_bytes,
            obj_name=self.obj_name,
            last_change=self.last_change
        )



import logging

from binsync.common.ui.qt_objects import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QVBoxLayout,
    QLabel,
    QComboBox,
    Qt
)


l = logging.getLogger(__name__)


class MagicSyncDialog(QDialog):
    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.controller = controller

        self.setWindowTitle("Magic Sync")
        self._main_layout = QGridLayout()
        self._init_widgets()
        self.setLayout(self._main_layout)

        self.should_sync = False
        self.preferred_user = None
        self.show()

    def _init_widgets(self):
        vertical_layout = QVBoxLayout()

        # dialog text
        self.label = QLabel(self)
        self.label.setWordWrap(True)
        self.label.setText(
            "<html><head/><body><p> Magic Sync is a one-time sync that attempts to sync <span style=\" "
            "font-weight:600;\">non-conflicting</span> data from all users on all functions, essentially a global "
            "knowledge merge. Would you like to preform this action? You may optionally select a user you would like "
            "prioritized for non-conflicting sync first.</p><p>Priority User:</p></body></html>",
        )

        # user selection
        items = self._get_users()
        self.comboBox = QComboBox(self)
        self.comboBox.addItems(items)

        # confirm button
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setStandardButtons(QDialogButtonBox.No | QDialogButtonBox.Yes)
        self.buttonBox.accepted.connect(self._on_yes_clicked)
        self.buttonBox.rejected.connect(self._on_no_clicked)

        vertical_layout.addWidget(self.label, 0, Qt.AlignBottom)
        vertical_layout.addWidget(self.comboBox, 0, Qt.AlignBottom)
        vertical_layout.addWidget(self.buttonBox, 0, Qt.AlignBottom)

        self._main_layout.addLayout(vertical_layout, 0, 0, 1, 1)

    def _get_users(self):
        return ["None"] + list(self.controller.usernames(priority=1))

    def _on_yes_clicked(self):
        self.should_sync = True
        combo_text = self.comboBox.currentText()
        self.preferred_user = combo_text if combo_text != "None" else None
        self.close()

    def _on_no_clicked(self):
        self.should_sync = False
        self.close()


def display_magic_sync_dialog(controller):
    dialog = MagicSyncDialog(controller)
    dialog.exec_()

    if not dialog.should_sync:
        return

    controller.magic_fill(preference_user=dialog.preferred_user)



import logging
import pathlib
import os
import re
import subprocess
import time
from functools import wraps
from typing import Iterable, Optional

import filelock
import git
import git.exc

from binsync.data import User, GlobalConfig
from binsync.core.errors import ExternalUserCommitError, MetadataNotFoundError
from binsync.data.state import State, load_toml_from_file
from binsync.core.scheduler import Scheduler, Job, SchedSpeed
from binsync.core.cache import Cache


l = logging.getLogger(__name__)
BINSYNC_BRANCH_PREFIX = 'binsync'
BINSYNC_ROOT_BRANCH = f'{BINSYNC_BRANCH_PREFIX}/__root__'


class ConnectionWarnings:
    HASH_MISMATCH = 0


def atomic_git_action(f):
    """
    Assures that any function called with this decorator will execute in-order, atomically, on a single thread.
    This all assumes that the function you are passing is a member of the Client class, which will also have
    a scheduler. This also means that this can only be called after the scheduler is started. This also requires a
    Cache. Generally, just never call functions with this decorator until the Client is done initing.

    This function will also attempt to check the cache for requested data on the same thread the original call
    was made from. If not found, the atomic scheduling is done.

    @param f:   A Client object function
    @return:
    """
    @wraps(f)
    def _atomic_git_action(self: "Client", *args, **kwargs):
        no_cache = kwargs.get("no_cache", False)
        if not no_cache:
            # cache check
            cache_item = self._check_cache_(f, **kwargs)
            if cache_item is not None:
                return cache_item

        # non cache available, queue it up!
        priority = kwargs.get("priority", None) or SchedSpeed.SLOW
        ret_val = self.scheduler.schedule_and_wait_job(
            Job(f, self, *args, **kwargs),
            priority=priority
        )

        self._set_cache(f, ret_val, **kwargs)

        return ret_val

    return _atomic_git_action


class Client:
    def __init__(
        self,
        master_user: str,
        repo_root: str,
        binary_hash: bytes,
        remote: str = "origin",
        commit_interval: int = 10,
        init_repo: bool = False,
        remote_url: Optional[str] = None,
        ssh_agent_pid: Optional[int] = None,
        ssh_auth_sock: Optional[str] = None
    ):
        """
        The Client class is responsible for making the low-level Git operations for the BinSync environment.
        Things like committing, pulling, and pushing all happen on the client level. It also starts a thread
        for continuous pushing.

        :param master_user:         Username of the user that is initing the client (the master user)
        :param repo_root:           Path to the BinSync repo where the project will be stored
        :param binary_hash:         The hash, usually md5, of the binary the client is connected for
        :param remote:              The optional Git remote (usually origin)
        :param commit_interval:     The seconds between each commit in the worker thread
        :param init_repo:           Bool to decide initing for both remote and local repos
        :param remote_url:          Remote URL to a Git Repo which may be used for cloning or initing
        :param ssh_agent_pid:       SSH Agent PID
        :param ssh_auth_sock:       SSH Auth Socket
        """
        self.master_user = master_user
        self.repo_root = repo_root
        self.binary_hash = binary_hash
        self.remote = remote
        self.repo = None
        self.repo_lock = None

        # validate this username can exist
        if master_user.endswith('/') or '__root__' in master_user:
            raise Exception(f"Bad username: {master_user}")

        # ssh-agent info
        self.ssh_agent_pid = ssh_agent_pid  # type: int
        self.ssh_auth_sock = ssh_auth_sock  # type: str
        self.connection_warnings = []

        # job scheduler
        self.cache = Cache()
        self.scheduler = Scheduler()

        # create, init, and checkout Git repo
        self.repo = self._get_or_init_binsync_repo(remote_url, init_repo)
        self.scheduler.start_worker_thread()
        self._get_or_init_user_branch()

        # timestamps
        self._commit_interval = commit_interval
        self._last_push_ts = None  # type: Optional[float]
        self.last_push_attempt_ts = None  # type: Optional[float]
        self._last_pull_ts = None  # type: Optional[float]
        self.last_pull_attempt_ts = None  # type: Optional[float]
        self._last_commit_ts = None # type: Optional[float]

        # load or update the global binsync config
        self.global_config = self._load_or_update_config()

        self.active_remote = True

    def __del__(self):
        if self.repo_lock is not None:
            self.repo_lock.release()

    #
    # Initializers
    #

    def _load_or_update_config(self):
        config = GlobalConfig.load_from_file(None) or GlobalConfig(None)
        config.last_bs_repo_path = self.repo_root
        config.save()
        return config

    def _get_or_init_user_branch(self):
        """
        Creates a user branch if the user is new, otherwise it gets the branch of the same
        name as the user

        @return:
        """
        try:
            branch = next(o for o in self.repo.branches if o.name.endswith(self.user_branch_name))
        except StopIteration:
            branch = self.repo.create_head(self.user_branch_name, BINSYNC_ROOT_BRANCH)
        else:
            if branch.is_remote():
                branch = self.repo.create_head(self.user_branch_name)
        branch.checkout()

    def _get_or_init_binsync_repo(self, remote_url, init_repo):
        """
        Gets the BinSync repo from either a local or remote git location and then sets up the repo as well
        as checks that the repo is the right repo for the current binary.

        When getting the repo there are four scenarios:
        1. The user fills out the remote_url and does not try to init it. In this case we should clone down
           the repo and assure that the head is remote/binsync/__root__
        2. The user fills out the remote_url and want to init it. In this case, we should clone it down to the repo
           root specified in the path. If there is no path, just place it locally then try to set it up with a hash
        3. The user fills out only the path of the git repo (no remote), then we just need to get that local git
           repo and assure it is a BinSync repo
        4. Last case is what happens if a user wants to init a local folder that is not a Git folder. In this case,
           they will be offline but will have a local git repo.

        @param remote_url:
        @param init_repo:
        @return:
        """
        if remote_url:
            # given a remove URL and no local folder, make it based on the URL name
            if not self.repo_root:
                self.repo_root = re.findall(r"/(.*)\.git", remote_url)[0]

            self.repo: git.Repo = self.clone(remote_url, no_head_check=init_repo)

            if init_repo:
                if any(b.name == BINSYNC_ROOT_BRANCH for b in self.repo.branches):
                    raise Exception("Can't init this remote repo since a BinSync root already exists")

                self._setup_repo()
        else:
            try:
                self.repo = git.Repo(self.repo_root)

                # update view of remote branches (which may be new)
                self._localize_remote_branches()

                if init_repo:
                    raise Exception("Could not initialize repository - it already exists!")
                if not any(b.name == BINSYNC_ROOT_BRANCH for b in self.repo.branches):
                    raise Exception(f"This is not a BinSync repo - it must have a {BINSYNC_ROOT_BRANCH} branch.")
            except (git.NoSuchPathError, git.InvalidGitRepositoryError):
                if init_repo:
                    # case 3
                    self.repo = git.Repo.init(self.repo_root)
                    self._setup_repo()
                else:
                    raise Exception(f"Failed to connect or create a BinSync repo")

        stored = self._get_stored_hash()
        if stored != self.binary_hash:
            self.connection_warnings.append(ConnectionWarnings.HASH_MISMATCH)

        assert not self.repo.bare, "it should not be a bare repo"

        self.repo_lock = filelock.FileLock(self.repo_root + "/.git/binsync.lock")
        try:
            self.repo_lock.acquire(timeout=0)
        except filelock.Timeout as e:
            raise Exception("Can only have one binsync client touching a local repository at once.\n"
                            "If the previous client crashed, you need to delete " + self.repo_root +
                            "/.git/binsync.lock") from e

        return self.repo

    def _setup_repo(self):
        """
        For use in initializing folder that is not yet a Git repo.

        @return:
        """
        with open(os.path.join(self.repo_root, ".gitignore"), "w") as f:
            f.write(".git/*\n")
        with open(os.path.join(self.repo_root, "binary_hash"), "w") as f:
            f.write(self.binary_hash)
        self.repo.index.add([".gitignore", "binary_hash"])
        self.repo.index.commit("Root commit")
        self.repo.create_head(BINSYNC_ROOT_BRANCH)

    #
    # Public Properties
    #

    @property
    def last_push_ts(self):
        return self._last_push_ts

    @property
    def last_pull_ts(self):
        return self._last_pull_ts

    @property
    def last_commit_ts(self):
        return self._last_commit_ts

    @property
    def user_branch_name(self):
        return f"{BINSYNC_BRANCH_PREFIX}/{self.master_user}"

    #
    # Atomic Public API
    #

    @atomic_git_action
    def commit_state(self, state, msg="Generic Change", priority=None):
        if self.master_user != state.user:
            raise ExternalUserCommitError(f"User {self.master_user} is not allowed to commit to user {state.user}")

        self._checkout_to_master_user()
        master_user_branch = next(o for o in self.repo.branches if o.name == self.user_branch_name)
        index = self.repo.index

        # dump the state
        state.dump(index)

        # commit changes
        self.repo.index.add([os.path.join(state.user, "*")])

        if not self.repo.index.diff("HEAD"):
            return

        # commit if there is any difference
        try:
            commit = index.commit(msg)
        except Exception as e:
            l.warning(f"Internal Git Commit Error: {e}")
            return
        self._last_commit_ts = time.time()
        master_user_branch.commit = commit
        state._dirty = False

    @atomic_git_action
    def users(self, priority=None, no_cache=False) -> Iterable[User]:
        repo = self.repo
        users = list()
        for ref in self._get_best_refs(repo).values():
            try:
                metadata = load_toml_from_file(ref.commit.tree, "metadata.toml", client=self)
                user = User.from_metadata(metadata)
                users.append(user)
            except Exception as e:
                l.debug(f"Unable to load user {e}")
                continue

        return users

    @atomic_git_action
    def get_state(self, user=None, version=None, priority=None, no_cache=False):
        if user is None:
            user = self.master_user

        repo = self.repo
        state = None
        try:
            state = State.parse(
                self._get_tree(user, repo),
                version=version,
                client=self
            )
        except MetadataNotFoundError:
            if user == self.master_user:
                state = State(self.master_user, client=self)

        return state

    @atomic_git_action
    def pull(self, priority=SchedSpeed.AVERAGE):
        """
        Pull changes from the remote side.

        :return:    None
        """
        self.last_pull_attempt_ts = time.time()
        try:
            env = self.ssh_agent_env()
        except Exception:
            return

        with self.repo.git.custom_environment(**env):
            # dangerous remote operations happen here
            try:
                self._localize_remote_branches()
                self.repo.git.checkout(BINSYNC_ROOT_BRANCH)
                self.repo.git.fetch("--all")
                self.repo.git.pull("--all")
                self._last_pull_ts = time.time()
                self.active_remote = True
            except Exception as e:
                l.debug(f"Pull exception {e}")
                self.active_remote = False

        if not self.active_remote:
            return

        # preform a merge on each branch
        for branch in self.repo.branches:
            if "HEAD" in branch.name:
                continue

            self.repo.git.checkout(branch)
            try:
                self.repo.git.merge()
            except Exception as e:
                #l.debug(f"Failed to merge on {branch} with {e}")
                pass

        self._update_cache()

    @atomic_git_action
    def push(self, print_error=False, priority=SchedSpeed.AVERAGE):
        """
        Push local changes to the remote side.

        :return:    None
        """
        self.last_push_attempt_ts = time.time()
        self._checkout_to_master_user()
        try:
            env = self.ssh_agent_env()
            with self.repo.git.custom_environment(**env):
                self.repo.remotes[self.remote].push(BINSYNC_ROOT_BRANCH)
                self.repo.remotes[self.remote].push(self.user_branch_name)
            self._last_push_ts = time.time()
            #l.debug("Push completed successfully at %s", self._last_push_ts)
            self.active_remote = True
        except git.exc.GitCommandError as ex:
            self.active_remote = False
            l.debug(f"Failed to push b/c {ex}")

    @property
    @atomic_git_action
    def has_remote(self, priority=SchedSpeed.FAST):
        """
        If there is a remote configured for our local repo.

        :return:    True if there is a remote, False otherwise.
        """
        return self.remote and any(r.name == self.remote for r in self.repo.remotes)

    #
    # Non-Atomic Public API
    #

    def update(self, commit_msg="Generic Change"):
        """
        Update both the local and remote repo knowledge of files through pushes/pulls and commits
        in the case of dirty files.
        """
        # attempt to commit dirty files in a update phase
        master_state = self.get_state(no_cache=True)
        if master_state and master_state.dirty:
            self.commit_state(master_state, msg=commit_msg)

        # do a pull if there is a remote repo connected
        self.last_pull_attempt_ts = time.time()
        if self.has_remote:
            self.pull()

        self.last_push_attempt_ts = time.time()
        if self.has_remote:
            self.push()

    #
    # Git Updates
    #

    def _localize_remote_branches(self):
        """
        Looks up all the remote refrences on the server and attempts to make them a tracked local
        branch.

        @return:
        """


        # get all remote branches
        try:
            remote_branches = self.repo.remote().refs
        except ValueError:
            return

        # track any remote we are not already tracking
        local_branches = set(b.name for b in self.repo.branches)
        for branch in remote_branches:
            # exclude head commit
            if "HEAD" in branch.name:
                continue

            # attempt to localize the remote name
            try:
                local_name = re.findall(f"({self.remote}/)(.*)", branch.name)[0][1]
            except IndexError:
                print("INDEX ERROR")
                continue

            # never try to track things already tracked
            if local_name in local_branches:
                continue

            try:
                self.repo.git.checkout('--track', branch.name)
            except git.GitCommandError as e:
                continue



    def ssh_agent_env(self):
        if self.ssh_agent_pid is not None and self.ssh_auth_sock is not None:
            env = {
                'SSH_AGENT_PID': str(self.ssh_agent_pid),
                'SSH_AUTH_SOCK': str(self.ssh_auth_sock),
            }
        else:
            env = { }
        return env

    def clone(self, remote_url, no_head_check=False):
        """
        Checkout from a remote_url to a local path specified by self.local_root.

        :param str remote_url:  The URL of the Git remote.
        :return:                None
        """

        env = self.ssh_agent_env()
        repo = git.Repo.clone_from(remote_url, self.repo_root, env=env)

        if no_head_check:
            return repo

        try:
            repo.create_head(BINSYNC_ROOT_BRANCH, f'{self.remote}/{BINSYNC_ROOT_BRANCH}')
        except git.BadName:
            raise Exception(f"This is not a binsync repo - it must have a {BINSYNC_ROOT_BRANCH} branch")

        return repo

    def _checkout_to_master_user(self):
        """
        Ensure the repo is in the proper branch for current user.

        :return: bool
        """
        self.repo.git.checkout(self.user_branch_name)

    @staticmethod
    def discover_ssh_agent(ssh_agent_cmd):
        proc = subprocess.Popen(ssh_agent_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()

        stdout = stdout.decode("utf-8")
        stderr = stderr.decode("utf-8")

        if proc.returncode != 0 or stderr:
            raise RuntimeError("Failed to discover SSH agent by running command %s.\n"
                               "Return code: %d.\n"
                               "stderr: %s" % (
                ssh_agent_cmd,
                proc.returncode,
                stderr,
            ))

        # parse output
        m = re.search(r"Found ssh-agent at (\d+)", stdout)
        if m is None:
            print("Failed to find 'Found ssh-agent at'")
            m = re.search(r"SSH_AGENT_PID=(\d+);", stdout)
            if m is None:
                print("Failed to find SSH_AGENT_PID")
                return None, None
            print("Found SSH_AGENT_PID")
            ssh_agent_pid = int(m.group(1))
            m = re.search("SSH_AUTH_SOCK=(.*?);", stdout)
            if m is None:
                print("Failed to find SSH_AUTH_SOCK")
                return None, None
            print("Found SSH_AGENT_SOCK")
            ssh_agent_sock = m.group(1)
        else :
            print("Found ssh-agent at")
            ssh_agent_pid = int(m.group(1))
            m = re.search(r"Found ssh-agent socket at ([^\s]+)", stdout)
            if m is None:
                print("Failed to find 'Found ssh-agent socket at'")
                return None, None
            print("Found ssh-agent socket at")
            ssh_agent_sock = m.group(1)

        return ssh_agent_pid, ssh_agent_sock

    def close(self):
        self.repo.close()
        del self.repo

    def _get_best_refs(self, repo):
        candidates = {}
        for ref in repo.refs:  # type: git.Reference
            if f'{BINSYNC_BRANCH_PREFIX}/' not in ref.name:
                continue

            branch_name = ref.name.split("/")[-1]
            if branch_name in candidates:
                # if the candidate exists, and the new one is not remote, don't replace it
                if not ref.is_remote() or ref.remote_name != self.remote:
                    continue

            candidates[branch_name] = ref
        return candidates

    def _get_stored_hash(self):
        branch = [ref for ref in self.repo.refs if ref.name.endswith(BINSYNC_ROOT_BRANCH)][0]
        return branch.commit.tree["binary_hash"].data_stream.read().decode().strip("\n")

    def list_files_in_tree(self, base_tree: git.Tree):
        """
        Lists all the files in a repo at a given tree

        :param commit: A gitpython Tree object
        """
        file_list = []
        stack = [base_tree]
        while len(stack) > 0:
            tree = stack.pop()
            # enumerate blobs (files) at this level
            for b in tree.blobs:
                file_list.append(b.path)
            for subtree in tree.trees:
                stack.append(subtree)

        return file_list

    def add_data(self, index: git.IndexFile, path: str, data: bytes):
        """
        Adds physical files to the database.

        WARNING: this function touches physical files in the Git Repo which can result in a race
        condition to modify a file while it is also being pushed. ONLY CALL THIS FUNCTION INSIDE
        A COMMIT_LOCK.

        @param index:
        @param path:
        @param data:
        @return:
        """
        fullpath = os.path.join(os.path.dirname(index.repo.git_dir), path)
        pathlib.Path(fullpath).parent.mkdir(parents=True, exist_ok=True)
        with open(fullpath, 'wb') as fp:
            fp.write(data)
        index.add([fullpath])

    def remove_data(self, index: git.IndexFile, path: str):
        fullpath = os.path.join(os.path.dirname(index.repo.git_dir), path)
        pathlib.Path(fullpath).parent.mkdir(parents=True, exist_ok=True)
        index.remove([fullpath], working_tree=True)

    def load_file_from_tree(self, tree: git.Tree, filename):
        return tree[filename].data_stream.read().decode()

    def _get_tree(self, user, repo: git.Repo):
        options = [ref for ref in repo.refs if ref.name.endswith(f"{BINSYNC_BRANCH_PREFIX}/{user}")]
        if not options:
            raise ValueError(f'No such user "{user}" found in repository')

        # find the latest commit for the specified user!
        best = max(options, key=lambda ref: ref.commit.authored_date)
        bct = best.commit.tree

        return bct

    #
    # Caching Functions
    #

    def _get_commits_for_users(self, repo: git.Repo):
        ref_dict = self._get_best_refs(repo)

        # ignore the _root_ branch
        if BINSYNC_ROOT_BRANCH in ref_dict:
            del ref_dict[BINSYNC_ROOT_BRANCH]

        commit_dict = {
            branch_name: ref.commit.hexsha for branch_name, ref in ref_dict.items()
        }
        return commit_dict

    def _check_cache_(self, f, **kwargs):
        if f.__qualname__ == self.get_state.__qualname__:
            cache_func = self.cache.get_state
            args = []
            if kwargs.get("user", None) is None:
                kwargs["user"] = self.master_user

        elif f.__qualname__ == self.users.__qualname__:
            cache_func = self.cache.users
            args = []
        else:
            return None

        item = cache_func(*args, **kwargs)

        return item

    def _set_cache(self, f, ret_value, **kwargs):
        if f.__qualname__ == self.get_state.__qualname__:
            set_func = self.cache.set_state
            args = []
            if kwargs.get("user", None) is None:
                kwargs["user"] = self.master_user
        elif f.__qualname__ == self.users.__qualname__:
            set_func = self.cache.set_users
            args = []
        else:
            return None

        set_func(ret_value, *args, **kwargs)

    def _update_cache(self):
        l.debug(f"Updating cache commits for State Cache...")
        cache_dict = self._get_commits_for_users(git.Repo(self.repo_root))
        self.cache.update_state_cache_commits(cache_dict)

        cache_keys = [key for key in cache_dict.keys()]
        l.debug(f"Updating branches on Users Cache...")
        branch_set = set(cache_keys)
        self.cache.update_user_cache_branches(branch_set)





import logging

from binsync.common.controller import BinSyncController, MergeLevel
from binsync.common.ui.qt_objects import (
    QCheckBox,
    QComboBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    Qt,
    QVBoxLayout,
    QWidget,
)
from binsync.common.ui.magic_sync_dialog import MagicSyncDialog
from binsync.common.ui.force_push import ForcePushUI
from binsync.common.controller import BinSyncController
from binsync.core.scheduler import SchedSpeed

l = logging.getLogger(__name__)


class QUtilPanel(QWidget):
    def __init__(self, controller: BinSyncController, parent=None):
        super().__init__(parent)
        self.controller = controller
        self._init_widgets()

    def _init_widgets(self):

        #
        # Developer Options Group
        #

        dev_options_group = QGroupBox()
        dev_options_layout = QVBoxLayout()
        dev_options_group.setTitle("Developer Options")
        dev_options_group.setLayout(dev_options_layout)

        self._debug_log_toggle = QCheckBox("Toggle Debug Logging")
        self._debug_log_toggle.setToolTip("Toggles the logging of events BinSync developers care about.")
        self._debug_log_toggle.stateChanged.connect(self._handle_debug_toggle)
        dev_options_layout.addWidget(self._debug_log_toggle)

        #
        # Sync Options Group
        #

        sync_options_group = QGroupBox()
        sync_options_layout = QVBoxLayout()
        sync_options_group.setTitle("Sync Options")
        sync_options_group.setLayout(sync_options_layout)

        self._merge_level_label = QLabel("Sync Merge Level")
        self._merge_level_label.setToolTip(
            """<html>
            <p>
            Defines which method is used to sync artifacts from another user.<br>
            <b>Non-Conflicting</b>: Only syncs artifacts that are not currently defined by you, so nothing is ever overwritten.<br>
            <b>Overwrite</b>: Syncs all artifacts regardless of your defined ones, overwriting everything.<br>
            <b>Merge</b>: You pick which artifacts are synced via the UI. <b>Unimplemented.</b>
            </p>
            </html>
            """)
        self._merge_level_label.setTextFormat(Qt.RichText)
        self._merge_level_combobox = QComboBox()
        self._merge_level_combobox.addItems(["Non-Conflicting", "Overwrite", "Merge"])
        self._merge_level_combobox.currentIndexChanged.connect(self._handle_sync_level_change)

        sync_level_layout = QHBoxLayout()
        #sync_level_group.layout().setContentsMargins(0, 0, 0, 0)
        sync_level_layout.addWidget(self._merge_level_label)
        sync_level_layout.addWidget(self._merge_level_combobox)

        self._magic_sync_button = QPushButton("Initiate Magic Sync")
        self._magic_sync_button.clicked.connect(self._handle_magic_sync_button)
        self._magic_sync_button.setToolTip("Performs a best effort merge of all existing user data to your state, "
                                           "but won't affect your existing state (this uses a non-conflicting merge).")

        self._force_push_button = QPushButton("Force Push...")
        self._force_push_button.clicked.connect(self._handle_force_push_button)
        self._force_push_button.setToolTip("Manually select function and globals you would like to be force committed "
                                           "and pushed to your user branch on Git.")

        sync_options_layout.addLayout(sync_level_layout)
        sync_options_group.layout().addWidget(self._magic_sync_button)
        sync_options_group.layout().addWidget(self._force_push_button)

        #
        # Final Layout
        #

        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(10, 20, 10, 20)
        main_layout.setSpacing(18)
        main_layout.setAlignment(Qt.AlignTop)
        main_layout.addWidget(sync_options_group)
        main_layout.addWidget(dev_options_group)
        self.setLayout(main_layout)

    #
    # Event Handlers
    #

    def _handle_debug_toggle(self):
        if self._debug_log_toggle.isChecked():
            logging.getLogger("binsync").setLevel("DEBUG")
            logging.getLogger("ida_binsync").setLevel("DEBUG")
            l.info("Logger has been set to level: DEBUG")
        else:
            logging.getLogger("binsync").setLevel("INFO")
            logging.getLogger("ida_binsync").setLevel("INFO")
            l.info("Logger has been set to level: INFO")

    def _handle_sync_level_change(self, index):
        selected_opt = self._merge_level_combobox.itemText(index)
        if selected_opt == "Non-Conflicting":
            self.controller.merge_level = MergeLevel.NON_CONFLICTING
        elif selected_opt == "Overwrite":
            self.controller.merge_level = MergeLevel.OVERWRITE
        elif selected_opt == "Merge":
            self.controller.merge_level = MergeLevel.MERGE
        else:
            return
        l.debug(f"Sync level changed to: {selected_opt}")

    def _handle_magic_sync_button(self):
        dialog = MagicSyncDialog(self.controller)
        dialog.exec_()

        if not dialog.should_sync:
            return

        self.controller.magic_fill(preference_user=dialog.preferred_user)

    def _handle_force_push_button(self):
        self.popup = ForcePushUI(self.controller)
        self.popup.show()



from typing import Dict

import toml


class Artifact:
    __slots__ = (
        "last_change",
    )

    def __init__(self, last_change=None):
        self.last_change = last_change

    def __getstate__(self) -> Dict:
        """
        Returns a dict of all the properties of the artifact. With the key as their name
        and the value as their value.

        @return:
        """
        return dict(
            (k, getattr(self, k)) for k in self.__slots__
        )

    def __setstate__(self, state):
        """
        Sets all the properties of the artifact given a dict of keys and values.
        Note: the values can also be dicts.

        @param state: Dict
        @return:
        """
        for k in self.__slots__:
            setattr(self, k, state.get(k, None))

    def __eq__(self, other):
        """
        Like a normal == override, but we always ignore last_push.

        @param other: Another Artifact
        @return:
        """
        if not isinstance(other, self.__class__):
            return False

        for k in self.__slots__:
            if k == "last_change":
                continue

            if getattr(self, k) != getattr(other, k):
                return False

        return True

    def diff(self, other, **kwargs) -> Dict:
        diff_dict = {}
        if not isinstance(other, self.__class__):
            for k in self.__slots__:
                if k == "last_change":
                    continue

                diff_dict[k] = {
                    "before": getattr(self, k),
                    "after": None
                }
            return diff_dict

        for k in self.__slots__:
            self_attr, other_attr = getattr(self, k), getattr(other, k)
            if self_attr != other_attr:
                if k == "last_change":
                    continue

                diff_dict[k] = {
                    "before": self_attr,
                    "after": other_attr
                }
        return diff_dict

    def dump(self) -> str:
        """
        Returns a string in TOML form of the properties of the current artifact. Best used to
        write directly into a file and save as a .toml file.

        @return:
        """
        return toml.dumps(self.__getstate__())

    def copy(self) -> "Artifact":
        return None

    @property
    def commit_msg(self) -> str:
        return f"Updated {self}"

    @classmethod
    def parse(cls, s):
        """
        Parses a TOML form string.

        @param s:
        @return:
        """
        raise NotImplementedError()

    @classmethod
    def invert_diff(cls, diff_dict: Dict):
        inverted_diff = {}
        for k, v in diff_dict.items():
            if k == "before":
                inverted_diff["after"] = v
            elif k == "after":
                inverted_diff["before"] = v
            elif isinstance(v, Dict):
                inverted_diff[k] = cls.invert_diff(v)
            else:
                inverted_diff[k] = v

        return inverted_diff

    def nonconflict_merge(self, obj2: "Artifact", **kwargs):
        obj1 = self.copy()
        if not obj2 or obj1 == obj2:
            return obj1

        obj_diff = obj1.diff(obj2)
        merge_obj = obj1.copy()

        for attr in self.__slots__:
            if attr in obj_diff and obj_diff[attr]["before"] is None:
                setattr(merge_obj, attr, getattr(obj2, attr))

        return merge_obj



import logging
from datetime import datetime

from binsync.common.controller import BinSyncController
from binsync.common.ui.qt_objects import (
    QAbstractItemView,
    QAbstractTableModel,
    QHeaderView,
    Qt,
    QModelIndex,
    QSortFilterProxyModel,
    QFocusEvent,
    QKeyEvent,
    QLineEdit,
    QTableView,
    QFontDatabase,
    QWidget,
    QVBoxLayout,
    QPushButton,
    QPersistentModelIndex,
    QCheckBox
)
from binsync.common.ui.utils import friendly_datetime
from binsync.data.state import State
from binsync.core.scheduler import SchedSpeed

l = logging.getLogger(__name__)


class FunctionTableModel(QAbstractTableModel):
    """Table model that controls backend behavior of the function table"""
    HEADER = [
        'Addr',
        'Remote Name',
        'User',
        'Last Push'
    ]

    # This is *most likely* alright, definitely works on linux, could use a macos/windows pass.
    # Custom defined role for sorting (since we shouldn't sort hex numbers alphabetically)
    SortRole = Qt.UserRole + 1000

    # Color for most recently updated, the alpha value decreases linearly over controller.table_coloring_window
    ACTIVE_FUNCTION_COLOR = (100, 255, 100, 70)

    def __init__(self, controller: BinSyncController, data=None, parent=None):
        super().__init__(parent)
        self.controller = controller
        self.row_data = data if data else []
        
        self.checks = [False for _ in self.row_data]
        self.HEADER[1] = 'Name'

    def rowCount(self, index=QModelIndex()):
        """ Returns number of rows the model holds. """
        return len(self.row_data)

    def columnCount(self, index=QModelIndex()):
        """ Returns number of columns the model holds. """
        return len(self.HEADER)

    def checkState(self, index):
        return Qt.Checked if self.checks[index.row()] else Qt.Unchecked

    def data(self, index, role=Qt.DisplayRole):
        """ Returns information about the data at a specified index based
            on the role supplied. """
        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            if index.column() == 0:
                if isinstance(self.row_data[index.row()][0], int):
                    return f"{self.row_data[index.row()][0]:#x}"
                else:
                    return self.row_data[index.row()][0]
            elif index.column() == 1:
                return self.row_data[index.row()][1]
            elif index.column() == 2:
                return self.row_data[index.row()][2]
            elif index.column() == 3:
                return friendly_datetime(self.row_data[index.row()][3])
        elif role == FunctionTableModel.SortRole:
            if index.column() == 0:
                return self.row_data[index.row()][0]
            elif index.column() == 1:
                return self.row_data[index.row()][1]
            elif index.column() == 2:
                return self.row_data[index.row()][2]
            elif index.column() == 3:
                return None
        elif role == Qt.CheckStateRole and index.column() == 0:
            return self.checkState(index)
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """ Set the headers to be displayed. """
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Horizontal:
            if 0 <= section < len(self.HEADER):
                return self.HEADER[section]

        return None

    def insertRows(self, position, rows=1, index=QModelIndex()):
        """ Insert N (default=1) rows into the model at a desired position. """
        self.beginInsertRows(QModelIndex(), position, position + rows - 1)

        for row in range(rows):
            self.row_data.insert(position + row, [0, "LOADING", "USER", datetime.now()])
            self.checks.insert(position+row, False)

        self.endInsertRows()
        return True

    def removeRows(self, position, rows=1, index=QModelIndex()):
        """ Remove N (default=1) rows from the model at a desired position. """
        if 0 <= position < len(self.row_data) and 0 <= position + rows < len(self.row_data):
            self.beginRemoveRows(QModelIndex(), position, position + rows - 1)

            del self.row_data[position:position + rows]
            self.endRemoveRows()
            return True
        return False

    def setData(self, index, value, role=Qt.EditRole):
        """ Adjust the data (set it to <value>) depending on the given
            index and role.
        """
        if role != Qt.EditRole and role != Qt.CheckStateRole:
            return False

        if index.isValid() and 0 <= index.row() < len(self.row_data):
            address = self.row_data[index.row()]
            if 0 == index.column() and role == Qt.CheckStateRole:
                self.checks[index.row()] = value
            elif 0 <= index.column() < len(address):
                address[index.column()] = value
            else:
                return False
            # TODO: Check for compatibility issues with pyqt6 here, function prototype changes between versions
            self.dataChanged.emit(index, index)
            return True

        return False

    def flags(self, index):
        """ Set the item flags at the given index. """
        if not index.isValid():
            return Qt.ItemIsEnabled
        if index.column() == 0:
            return Qt.ItemIsUserCheckable | Qt.ItemIsSelectable | Qt.ItemIsEnabled
        else:
            return Qt.ItemFlags(QAbstractTableModel.flags(self, index))

    def update_table(self):
        for address, function in self.controller.functions().items():
            self.insertRows(0)
            row_data = [address, function.name, "", -1]
            for i in range(len(self.HEADER)):
                idx = self.index(0, i, QModelIndex())
                self.setData(idx, row_data[i], role=Qt.EditRole)


class FunctionTableFilterLineEdit(QLineEdit):
    """ Basic class for the filter line edit, clears itself whenever focus is lost. """
    def __init__(self, parent=None):
        super(FunctionTableFilterLineEdit, self).__init__(parent=parent)
        self.user_unfocused = False

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if self.user_unfocused:
            self.user_unfocused = False
            self.clear()

        if event.key() == Qt.Key_Escape:
            self.clear()
            return
        super(FunctionTableFilterLineEdit, self).keyPressEvent(event)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        if event.reason() == Qt.MouseFocusReason:
            self.user_unfocused = True
        super(FunctionTableFilterLineEdit, self).focusOutEvent(event)


class FunctionTableView(QTableView):
    """ Table view for the data, this is the front end "container" for our model. """

    def __init__(self, controller: BinSyncController, filteredit: FunctionTableFilterLineEdit, parent=None):
        super().__init__(parent=parent)
        self.controller = controller

        self.filteredit = filteredit
        self.filteredit.textChanged.connect(self.handle_filteredit_change)

        # Create a SortFilterProxyModel to allow for sorting/filtering
        self.proxymodel = QSortFilterProxyModel()
        # Set the sort role/column to filter by
        self.proxymodel.setSortRole(FunctionTableModel.SortRole)
        self.proxymodel.setFilterRole(Qt.DisplayRole)
        self.proxymodel.setFilterKeyColumn(-1)

        # Connect our model to the proxy model
        self.model = FunctionTableModel(controller)
        self.proxymodel.setSourceModel(self.model)
        self.setModel(self.proxymodel)
        self.column_visibility = []

        self._init_settings()

    def _doubleclick_handler(self):
        """ Handler for double clicking on a row, jumps to the respective function. """
        row_idx = self.selectionModel().selectedIndexes()[0]
        tls_row_idx = self.proxymodel.mapToSource(row_idx)
        row = self.model.row_data[tls_row_idx.row()]
        self.controller.goto_address(row[0])

    def _get_valid_users_for_func(self, func_addr):
        """ Helper function for getting users that have changes in a given function """
        for user in self.controller.users(priority=SchedSpeed.FAST):
            user_state: State = self.controller.client.get_state(user=user.name, priority=SchedSpeed.FAST)
            user_func = user_state.get_function(func_addr)

            # function must be changed by this user
            if not user_func or not user_func.last_change:
                continue

            yield user.name

    def _col_hide_handler(self, index):
        """ Helper function to hide/show columns from context menu """
        self.column_visibility[index] = not self.column_visibility[index]
        self.setColumnHidden(index, self.column_visibility[index])
        if self.column_visibility[index]:
            self.showColumn(index)
        else:
            self.hideColumn(index)

    def update_table(self):
        """ Update the model of the table with new data from the controller """
        self.model.update_table()

    def _init_settings(self):
        self.setShowGrid(False)

        header = self.horizontalHeader()
        header.setSortIndicator(0, Qt.AscendingOrder)
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(3, QHeaderView.ResizeToContents)

        self.column_visibility = [True for _ in range(self.model.columnCount())]

        fixed_width_font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        fixed_width_font.setPointSize(11)
        self.setFont(fixed_width_font)

        self.setSortingEnabled(True)
        self.setSelectionMode(QAbstractItemView.NoSelection)

        self.setEditTriggers(QAbstractItemView.NoEditTriggers)

        self.setWordWrap(False)

        vheader = self.verticalHeader()
        vheader.setDefaultSectionSize(24)
        vheader.hide()

        self.setFocusProxy(self.filteredit)

    def handle_filteredit_change(self, text):
        """ Handle text changes in the filter box, filters the table by the arg. """
        self.proxymodel.setFilterFixedString(text)
        self.select_all.setChecked(False)
        for i in range(self.proxymodel.rowCount()):
            proxyIndex = self.proxymodel.index(i, 0, QModelIndex())
            mappedIndex = self.proxymodel.mapToSource(proxyIndex)
            if not self.model.data(mappedIndex, Qt.CheckStateRole):
                break
        else:
            self.select_all.setChecked(True)

    def push(self):
        self.proxymodel.setFilterFixedString("")
        for i in range(self.proxymodel.rowCount()):
            proxyIndex = self.proxymodel.index(i, 0, QModelIndex())
            mappedIndex = self.proxymodel.mapToSource(proxyIndex)
            if self.model.checkState(mappedIndex):
                func_addr = int(self.model.data(mappedIndex), 16)
                self.controller.force_push_function(func_addr)

    def connect_select_all(self, checkbox):
        self.select_all = checkbox

    def check_all(self):
        for i in range(self.proxymodel.rowCount()):
            proxyIndex = self.proxymodel.index(i, 0, QModelIndex())
            mappedIndex = self.proxymodel.mapToSource(proxyIndex)
            self.model.setData(mappedIndex, True, Qt.CheckStateRole)

    def uncheck_all(self):
        for i in range(self.proxymodel.rowCount()):
            proxyIndex = self.proxymodel.index(i, 0, QModelIndex())
            mappedIndex = self.proxymodel.mapToSource(proxyIndex)
            self.model.setData(mappedIndex, False, Qt.CheckStateRole)

class QFunctionTable(QWidget):
    """ Wrapper widget to contain the function table classes in one file (prevents bulking up control_panel.py) """
    def __init__(self, controller: BinSyncController, parent=None):
        super().__init__(parent)
        self.controller = controller
        self._init_widgets()

    def toggle_select_all(self):
        if self.checkbox.isChecked():
            self.table.check_all()
        else:
            self.table.uncheck_all()

    def _init_widgets(self):
        self.filteredit = FunctionTableFilterLineEdit(parent=self)
        self.table = FunctionTableView(self.controller, self.filteredit, parent=self)
        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        
        self.checkbox = QCheckBox("Select All")
        self.checkbox.clicked.connect(self.toggle_select_all)
        self.table.connect_select_all(self.checkbox)
        layout.addWidget(self.checkbox)
        layout.addWidget(self.table)
        layout.addWidget(self.filteredit)
        self.push_button = QPushButton("Push")
        self.push_button.clicked.connect(self.table.push)
        layout.addWidget(self.push_button)

        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def update_table(self):
        self.table.update_table()



import toml

from binsync.data.artifact import Artifact


class GlobalVariable(Artifact):
    __slots__ = Artifact.__slots__ + (
        "addr",
        "name",
        "type_str",
        "size"
    )

    def __init__(self, addr, name, type_str=None, size=0, last_change=None):
        super(GlobalVariable, self).__init__(last_change=last_change)
        self.addr = addr
        self.name = name
        self.type_str = type_str
        self.size = size

    def __str__(self):
        return f"<GlobalVar: {self.type_str} {self.name}; @{self.addr} len={self.size}>"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def parse(cls, s):
        gv = GlobalVariable(None, None)
        gv.__setstate__(toml.loads(s))
        return gv

    @classmethod
    def load_many(cls, gvars_toml):
        for gvar_toml in gvars_toml.values():
            global_var = GlobalVariable(None, None)
            try:
                global_var.__setstate__(gvar_toml)
            except TypeError:
                # skip all incorrect ones
                continue
            yield global_var

    @classmethod
    def dump_many(cls, global_vars):
        global_vars_ = {}

        for v in sorted(global_vars.values(), key=lambda x: x.addr):
            global_vars_["%x" % v.addr] = v.__getstate__()
        return global_vars_

    def copy(self):
        gvar = GlobalVariable(self.addr, self.name, self.type_str, self.size)
        return gvar



from typing import Dict, List

import toml

from binsync.data.artifact import Artifact

import logging
l = logging.getLogger(name=__name__)

class StructMember(Artifact):
    """
    Describes a struct member that corresponds to a struct.
    """

    __slots__ = (
        "last_change",
        "member_name",
        "offset",
        "type",
        "size",
    )

    def __init__(self, member_name, offset, type_, size, last_change=None):
        super(StructMember, self).__init__(last_change=last_change)
        self.member_name: str = member_name
        self.offset: int = offset
        self.type: str = type_
        self.size: int = size

    def __str__(self):
        return f"<StructMember: {self.type} {self.member_name}; @{hex(self.offset)}>"

    def __repr__(self):
        self.__str__()

    @classmethod
    def parse(cls, s):
        sm = StructMember(None, None, None, None)
        sm.__setstate__(toml.loads(s))
        return sm

    def copy(self):
        sm = StructMember(
            self.member_name,
            self.offset,
            self.type,
            self.size
        )

        return sm


class Struct(Artifact):
    """
    Describes a struct
    """

    __slots__ = (
        "last_change",
        "name",
        "size",
        "struct_members",
    )

    def __init__(self, name: str, size: int, struct_members: Dict[int, StructMember], last_change=None):
        super(Struct, self).__init__(last_change=last_change)
        self.name = name
        self.size = size or 0
        self.struct_members: Dict[int, StructMember] = struct_members

    def __str__(self):
        return f"<Struct: {self.name} membs={len(self.struct_members)} ({hex(self.size)})>"

    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        return {
            "metadata": {
                "name": self.name, "size": self.size, "last_change": self.last_change
            },

            "members": {
                "%x" % offset: member.__getstate__() for offset, member in self.struct_members.items()
            }
        }

    def __setstate__(self, state):
        metadata = state["metadata"]
        members = state["members"]

        self.name = metadata["name"]
        self.size = metadata["size"]
        self.last_change = metadata.get("last_change", None)

        self.struct_members = {
            int(off, 16): StructMember.parse(toml.dumps(member)) for off, member in members.items()
        }

    def add_struct_member(self, mname, moff, mtype, size):
        self.struct_members[moff] = StructMember(mname, moff, mtype, size)

    def diff(self, other, **kwargs) -> Dict:
        diff_dict = {}
        if not isinstance(other, Struct):
            return diff_dict

        for k in ["name", "size"]:
            if getattr(self, k) == getattr(other, k):
                continue

            diff_dict[k] = {
                "before": getattr(self, k),
                "after": getattr(other, k)
            }

        # struct members
        diff_dict["struct_members"] = {}
        for off, member in self.struct_members.items():
            try:
                other_mem = other.struct_members[off]
            except KeyError:
                other_mem = None

            diff_dict["struct_members"][off] = member.diff(other_mem)

        for off, other_mem in other.struct_members.items():
            if off in diff_dict["struct_members"]:
                continue

            diff_dict["struct_members"][off] = self.invert_diff(other_mem.diff(None))

        return diff_dict

    def copy(self):
        struct_members = {offset: member.copy() for offset, member in self.struct_members.items()}
        struct = Struct(self.name, self.size, struct_members, last_change=self.last_change)
        return struct

    @classmethod
    def parse(cls, s):
        struct = Struct(None, None, None)
        struct.__setstate__(s)
        return struct

    @classmethod
    def load(cls, struct_toml):
        s = Struct(None, None, None)
        s.__setstate__(struct_toml)
        return s

    def nonconflict_merge(self, struct2: "Struct", **kwargs) -> "Struct":
        struct1: "Struct" = self.copy()
        if not struct2 or struct1 == struct2:
            return struct1

        struct_diff = struct1.diff(struct2)
        merge_struct = struct1

        members_diff = struct_diff["struct_members"]
        for off, mem in struct2.struct_members.items():
            # no difference
            if off not in members_diff:
                continue

            mem_diff = members_diff[off]

            # struct member is newly created
            if "before" in mem_diff and mem_diff["before"] is None:
                # check for overlap
                new_mem_size = mem.size
                new_mem_offset = mem.offset

                for off_check in range(new_mem_offset, new_mem_offset + new_mem_size):
                    if off_check in merge_struct.struct_members:
                        break
                else:
                    merge_struct.struct_members[off] = mem.copy()

                continue

            # member differs
            merge_mem = merge_struct.struct_members.get(off, None)
            if not merge_mem:
                merge_mem = mem

            merge_mem = StructMember.nonconflict_merge(merge_mem, mem)
            merge_struct.struct_members[off] = merge_mem

        # compute the new size
        merge_struct.size = sum(mem.size for mem in merge_struct.struct_members.values())
        return merge_struct



import logging
import os
from typing import Optional, Dict

import angr
import binsync
from angr.analyses.decompiler.structured_codegen import DummyStructuredCodeGenerator
from angrmanagement.ui.views import CodeView

from binsync import Function
from binsync.common.controller import (
    BinSyncController,
    init_checker,
    fill_event
)
from binsync.data import (
    FunctionHeader, StackOffsetType, Comment, StackVariable
)

from .artifact_lifter import AngrArtifactLifter

l = logging.getLogger(__name__)


class AngrBinSyncController(BinSyncController):
    """
    The class used for all pushing/pulling and merging based actions with BinSync data.
    This class is responsible for handling callbacks that are done by changes from the local user
    and responsible for running a thread to get new changes from other users.
    """

    def __init__(self, workspace):
        super(AngrBinSyncController, self).__init__(artifact_lifter=AngrArtifactLifter(self))
        self._workspace = workspace
        self._instance = workspace.instance

    def binary_hash(self) -> str:
        return self._instance.project.loader.main_object.md5.hex()

    def active_context(self):
        curr_view = self._workspace.view_manager.current_tab
        if not curr_view:
            return None

        try:
            func = curr_view.function
        except NotImplementedError:
            return None

        if func is None or func.am_obj is None:
            return None

        func_addr = self.rebase_addr(func.addr)

        return binsync.data.Function(
            func_addr, 0, header=FunctionHeader(func.name, func_addr)
        )

    def binary_path(self) -> Optional[str]:
        try:
            return self._instance.project.loader.main_object.binary
        # pylint: disable=broad-except
        except Exception:
            return None

    def get_func_size(self, func_addr) -> int:
        try:
            func = self._instance.kb.functions[func_addr]
            return func.size
        except KeyError:
            return 0

    def rebase_addr(self, addr, up=False):
        base_addr = self._instance.project.loader.main_object.mapped_base
        is_pie = self._instance.project.loader.main_object.pic

        if is_pie:
            if up:
                return addr + base_addr
            elif addr > base_addr:
                return addr - base_addr

        return addr

    #
    # Display Fillers
    #
    @init_checker
    def fill_global_var(self, var_addr, user=None, artifact=None, **kwargs):
        return False

    @init_checker
    def fill_struct(self, struct_name, user=None, artifact=None, **kwargs):
        return False

    @init_checker
    @fill_event
    def fill_function(self, func_addr, user=None, artifact=None, **kwargs):
        func: Function = artifact
        func = self._instance.kb.functions[func.addr]

        # re-decompile a function if needed
        decompilation = self.decompile_function(func)

        changes = super(AngrBinSyncController, self).fill_function(
            func_addr, user=user, artifact=artifact, decompilation=decompilation, **kwargs
        )

        decompilation.regenerate_text()
        self.decompile_function(func, refresh_gui=True)
        return changes

    @init_checker
    @fill_event
    def fill_comment(self, addr, user=None, artifact=None, decompilation=None, **kwargs):
        cmt: Comment = artifact
        changed = False

        if cmt.decompiled:
            try:
                pos = decompilation.map_addr_to_pos.get_nearest_pos(addr)
                corrected_addr = decompilation.map_pos_to_addr.get_node(pos).tags['ins_addr']
            # pylint: disable=broad-except
            except Exception:
                return False

            if decompilation.stmt_comments[corrected_addr] != cmt.comment:
                decompilation.stmt_comments[corrected_addr] = cmt.comment
                changed = True
        else:
            if self._instance.kb.comments[cmt.addr] != cmt.comment:
                self._instance.kb.comments[cmt.addr] = cmt.comment
                changed = True
        return changed

    @init_checker
    @fill_event
    def fill_stack_variable(self, func_addr, offset, user=None, artifact=None, decompilation=None, **kwargs):
        sync_var: StackVariable = artifact
        changed = False
        code_var = AngrBinSyncController.find_stack_var_in_codegen(decompilation, offset)
        if code_var:
            code_var.name = sync_var.name
            code_var.renamed = True
            changed = True

        return changed

    @init_checker
    @fill_event
    def fill_function_header(self, func_addr, user=None, artifact=None, decompilation=None, **kwargs):
        func_header: FunctionHeader = artifact
        angr_func = self._instance.kb.functions[self.artifact_lifer.lower_addr(func_addr)]
        changes = False
        if func_header:
            if func_header.name and func_header.name != angr_func.name:
                angr_func.name = func_header.name
                decompilation.cfunc.name = func_header.name
                decompilation.cfunc.demangled_name = func_header.name
                changes = True

            if func_header.args:
                for i, arg in func_header.args.items():
                    if i >= len(decompilation.cfunc.arg_list):
                        break
                    if decompilation.cfunc.args_list[i].variable.name != arg.name:
                        decompilation.cfunc.arg_list[i].variable.name = arg.name
                        changes = True
        return changes

    #
    # Artifact
    #

    def functions(self) -> Dict[int, Function]:
        funcs = {}
        for addr, func in self._instance.kb.functions.items():
            funcs[addr] = Function(addr, func.size)
            funcs[addr].name = func.name

        return funcs

    def function(self, addr) -> Optional[Function]:
        """
        TODO: add support for stack variables and function args

        @param addr:
        @return:
        """
        try:
            _func = self._instance.kb.functions[addr]
        except KeyError:
            return None

        func = Function(_func.addr, _func.size)
        func_header = FunctionHeader(_func.name, _func.addr, ret_type=_func.prototype.c_repr())

        func.header = func_header
        return func


    #
    #   Utils
    #

    def decompile_function(self, func, refresh_gui=False):
        # check for known decompilation
        available = self._instance.kb.structured_code.available_flavors(func.addr)
        should_decompile = False
        if 'pseudocode' not in available:
            should_decompile = True
        else:
            cached = self._instance.kb.structured_code[(func.addr, 'pseudocode')]
            if isinstance(cached, DummyStructuredCodeGenerator):
                should_decompile = True

        if should_decompile:
            # recover direct pseudocode
            self._instance.project.analyses.Decompiler(func, flavor='pseudocode')

            # attempt to get source code if its available
            source_root = None
            if self._instance.original_binary_path:
                source_root = os.path.dirname(self._instance.original_binary_path)
            self._instance.project.analyses.ImportSourceCode(func, flavor='source', source_root=source_root)

        # grab newly cached pseudocode
        decomp = self._instance.kb.structured_code[(func.addr, 'pseudocode')].codegen
        if refresh_gui:
            # refresh all views
            self._workspace.reload()

            # re-decompile current view to cause a refresh
            current_tab = self._workspace.view_manager.current_tab
            if isinstance(current_tab, CodeView) and current_tab.function == func:
                self._workspace.decompile_current_function()

        return decomp

    @staticmethod
    def find_stack_var_in_codegen(decompilation, stack_offset: int) -> Optional[angr.sim_variable.SimStackVariable]:
        for var in decompilation.cfunc.variable_manager._unified_variables:
            if hasattr(var, "offset") and var.offset == stack_offset:
                return var

        return None

    @staticmethod
    def stack_var_type_str(decompilation, stack_var: angr.sim_variable.SimStackVariable):
        try:
            var_type = decompilation.cfunc.variable_manager.get_variable_type(stack_var)
        # pylint: disable=broad-except
        except Exception:
            return None

        return var_type.c_repr()

    @staticmethod
    def get_func_args(decompilation):
        arg_info = {
            i: (arg.variable, decompilation.cfunc.functy.args[i].c_repr())
            for i, arg in enumerate(decompilation.cfunc.arg_list)
        }
        return arg_info

    @staticmethod
    def func_insn_addrs(func: angr.knowledge_plugins.Function):
        insn_addrs = set()
        for block in func.blocks:
            insn_addrs.update(block.instruction_addrs)

        return insn_addrs

    def get_func_addr_from_addr(self, addr):
        try:
            func_addr = self._workspace.instance.kb.cfgs.get_most_accurate()\
                .get_any_node(addr, anyaddr=True)\
                .function_address
        except AttributeError:
            func_addr = None

        return func_addr
    
    def goto_address(self, func_addr):
        self._workspace.jump_to(self.rebase_addr(func_addr, up=True))



import xmlrpc.client
from functools import wraps

import json
from typing import Optional

from binsync.data import (
    Function, FunctionHeader
)


def stringify_args(f):
    @wraps(f)
    def _stringify_args(self, *args, **kwargs):
        new_args = list()
        for arg in args:
            if isinstance(arg, int) and not isinstance(arg, bool):
                new_arg = hex(arg)
            elif isinstance(arg, dict):
                new_arg = json.loads(json.dumps(arg), parse_int=lambda o: hex(o))
            else:
                new_arg = arg

            new_args.append(new_arg)

        return f(self, *new_args, **kwargs)

    return _stringify_args

class BSGhidraClient:
    def __init__(self, host="localhost", port=6683):
        self.host = host
        self.port = port

        self.server = None

    #
    # Server Ops
    #

    @property
    def connected(self):
        return True if self.server else False

    def connect(self, host=None, port=None) -> bool:
        """
        Connects to the remote decompiler.
        """
        host = host or self.host
        port = port or self.port

        # create a server connection and test it
        try:
            self.server = xmlrpc.client.ServerProxy(f"http://{host}:{port}").bs
            self.server.ping()
        except (ConnectionRefusedError, AttributeError) as e:
            self.server = None
            return False

        return True

    def alert_ui_configured(self, status):
        self.server.alertUIConfigured(status)

    #
    # Public Facing API
    #

    def context(self):
        if not self.server:
            return Function(0, 0, header=FunctionHeader("", 0))

        out = self.server.context()
        name = out["name"] or ""
        try:
            addr = int(out["addr"], 16)
        except:
            addr = 0

        return Function(addr, 0, header=FunctionHeader(name, addr))

    @property
    def base_addr(self) -> Optional[int]:
        val = self.server.baseAddr()
        if not val:
            return None

        return int(val, 16)

    @property
    def binary_hash(self) -> str:
        return self.server.binaryHash()

    @property
    def binary_path(self) -> Optional[str]:
        return self.server.binaryPath()

    @stringify_args
    def goto_address(self, addr) -> bool:
        return self.server.gotoAddress(addr)

    #
    # Function Operations
    #

    @stringify_args
    def set_func_name(self, addr: int, name: str) -> bool:
        return self.server.setFunctionName(addr, name)

    @stringify_args
    def set_func_rettype(self, addr: int, type_str: str) -> bool:
        return self.server.setFunctionRetType(addr, type_str)

    @stringify_args
    def set_stack_var_name(self, addr: int, offset: int, name: str) -> bool:
        return self.server.setStackVarName(addr, offset, name)

    @stringify_args
    def set_stack_var_type(self, addr: int, offset: int, type_: str) -> bool:
        return self.server.setStackVarType(addr, offset, type_)

    #
    # Comment Ops
    #

    @stringify_args
    def set_comment(self, addr: int, comment: str, is_decompiled: bool) -> bool:
        return self.server.setComment(addr, comment, is_decompiled)



import datetime
import logging

from binsync.common.ui.qt_objects import Qt, QTableWidgetItem

l = logging.getLogger(__name__)

class QNumericItem(QTableWidgetItem):
    def __lt__(self, other):
        if self.data(Qt.UserRole) is None:
            return True
        elif other.data(Qt.UserRole) is None:
            return False

        return self.data(Qt.UserRole) < other.data(Qt.UserRole)


def friendly_datetime(time_before):
    # convert fro unix
    if isinstance(time_before, int):
        if time_before == -1:
            return ""
        dt = datetime.datetime.fromtimestamp(time_before)
    elif isinstance(time_before, datetime.datetime):
        dt = time_before
    else:
        return ""

    now = datetime.datetime.now()
    if dt <= now:
        diff = now - dt
        ago = True
    else:
        diff = dt - now
        ago = False
    diff_days = diff.days
    diff_sec = diff.seconds

    if diff_days >= 1:
        s = "%d days" % diff_days
    elif diff_sec >= 60 * 60:
        s = "%d hours" % int(diff_sec / 60 / 60)
    elif diff_sec >= 60:
        s = "%d minutes" % int(diff_sec / 60)
    else:
        s = "%d seconds" % diff_sec

    s += " ago" if ago else " in the future"
    return s

def menu_stub(menu):
    return menu



import threading
import logging
from queue import PriorityQueue
from threading import Thread

l = logging.getLogger(__name__)


class SchedSpeed:
    FAST = 1
    AVERAGE = 2
    SLOW = 3


class FailedJob:
    def __init__(self, reason):
        self.reason = reason


class Job:
    def __init__(self, function, *args, **kwargs):
        self.function = function
        self.args = args
        self.kwargs = kwargs

        self.ret_value = None
        self.finish_event = threading.Event()

    def execute(self):
        self.ret_value = self.function(*self.args, **self.kwargs)
        self.finish_event.set()

    def __str__(self):
        return f"<Job: {self.function}({self.args}, {self.kwargs})>"

    def __repr__(self):
        return self.__str__()

    def __lt__(self, other):
        return True


class Scheduler:
    def __init__(self, sleep_interval=0.05):
        self.sleep_interval = sleep_interval
        self._worker = Thread(target=self._worker_thread)
        self._job_queue = PriorityQueue()
        self._work = False

    def stop_worker_thread(self):
        self._work = False

    def start_worker_thread(self):
        self._work = True
        self._worker.setDaemon(True)
        self._worker.start()

    def _worker_thread(self):
        while self._work:
            self._complete_a_job(block=True)

    def schedule_job(self, job: Job, priority=SchedSpeed.SLOW):
        self._job_queue.put_nowait(
            (priority, job,)
        )

    def schedule_and_wait_job(self, job: Job, priority=SchedSpeed.SLOW, timeout=30):
        self.schedule_job(job, priority=priority)
        try:
            job.finish_event.wait(timeout=timeout)
        except Exception as e:
            return FailedJob(e)
        else:
            return job.ret_value

    def _complete_a_job(self, block=False):
        if block:
            _, job = self._job_queue.get()
        elif self._job_queue.not_empty:
            _, job = self._job_queue.get_nowait()
        else:
            return

        job.execute()



import logging
import re

from binsync.common.controller import BinSyncController
from binsync.common.ui.qt_objects import (
    QAbstractItemView,
    QHeaderView,
    QMenu,
    Qt,
    QTableWidget,
    QTableWidgetItem,
)
from binsync.common.ui.utils import QNumericItem, friendly_datetime
from binsync.data.state import State
from binsync.core.scheduler import SchedSpeed

l = logging.getLogger(__name__)

class QGlobalItem:
    def __init__(self, name, type_, user, last_push):
        self.name = name
        self.type = type_
        self.user = user
        self.last_push = last_push

    def widgets(self):
        # sort by int value
        name = QTableWidgetItem(self.name)
        type_ = QTableWidgetItem(self.type)
        user = QTableWidgetItem(self.user)

        # sort by unix value
        last_push = QNumericItem(friendly_datetime(self.last_push))
        last_push.setData(Qt.UserRole, self.last_push)

        widgets = [
            name,
            type_,
            user,
            last_push
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QGlobalsTable(QTableWidget):
    HEADER = [
        'Name',
        'Type',
        'User',
        'Last Push'
    ]

    def __init__(self, controller: BinSyncController, parent=None):
        super(QGlobalsTable, self).__init__(parent)
        self.controller = controller
        self.items = []

        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.horizontalHeader().setHorizontalScrollMode(self.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignHCenter | Qt.Alignment(Qt.TextWordWrap))
        self.horizontalHeader().setMinimumWidth(160)
        self.setHorizontalScrollMode(self.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self.setSortingEnabled(True)

    def reload(self):
        self.setSortingEnabled(False)
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()
        self.setSortingEnabled(True)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        menu.setObjectName("binsync_global_table_context_menu")

        # create a nested menu
        selected_row = self.rowAt(event.pos().y())
        item0 = self.item(selected_row, 0)
        item1 = self.item(selected_row, 1)
        item2 = self.item(selected_row, 2)
        if any(x is None for x in [item0, item1, item2]):
            return
        global_name = item0.text()
        global_type = item1.text()
        user_name = item2.text()

        if global_type == "Struct":
            filler_func = self.controller.fill_struct
        elif global_type == "Variable":
            global_name = int(re.findall(r'0x[a-f,0-9]+', global_name.split(" ")[1])[0], 16)
            filler_func = self.controller.fill_global_var
        elif global_type == "Enum":
            filler_func = self.controller.fill_enum
        else:
            l.warning(f"Invalid global table sync option: {global_type}")
            return

        menu.addAction("Sync", lambda: filler_func(global_name, user=username))
        from_menu = menu.addMenu("Sync from...")
        for username in self._get_valid_users_for_global(global_name, global_type):
            action = from_menu.addAction(username)
            action.triggered.connect(lambda chck, name=username: filler_func(global_name, user=name))

        menu.popup(self.mapToGlobal(event.pos()))

    def update_table(self):
        known_globals = {}

        for user in self.controller.users():
            state = self.controller.client.get_state(user=user.name)
            user_structs = state.structs
            user_gvars = state.global_vars
            user_enums = state.enums

            all_artifacts = ((user_enums, "Enum"), (user_structs, "Struct"), (user_gvars, "Variable"))
            for user_artifacts, global_type in all_artifacts:
                for _, artifact in user_artifacts.items():
                    change_time = artifact.last_change

                    if not change_time:
                        continue

                    if artifact.name in known_globals:
                        # change_time < artifact_stored_change_time
                        if not change_time or change_time < known_globals[artifact.name][3]:
                            continue

                    artifact_name = artifact.name if global_type != "Variable" \
                        else f"{artifact.name} ({hex(artifact.addr)})"

                    known_globals[artifact_name] = (artifact_name, global_type, user.name, change_time)

        self.items = [QGlobalItem(*row) for row in known_globals.values()]

    def _get_valid_users_for_global(self, global_name, global_type):
        if global_type == "Struct":
            global_getter = "get_struct"
        elif global_type == "Variable":
            global_getter = "get_global_var"
        elif global_type == "Enum":
            global_getter = "get_enum"
        else:
            l.warning("Failed to get a valid type for global type")
            return

        for user in self.controller.users(priority=SchedSpeed.FAST):
            user_state: State = self.controller.client.get_state(user=user.name, priority=SchedSpeed.FAST)
            get_global = getattr(user_state, global_getter)
            user_global = get_global(global_name)

            # function must be changed by this user
            if not user_global or not user_global.last_change:
                continue

            yield user.name



from binsync.common import ArtifactLifter


class BinjaArtifactLifter(ArtifactLifter):
    lift_map = {
        "int64_t": "long",
        "uint64_t": "unsigned long",
        "int32_t": "int",
        "uint32_t": "unsigned int",
        "int16_t": "short",
        "uint16_t": "unsigned short",
        "int8_t": "char",
        "uint8_t": "unsigned char",
    }

    def __init__(self, controller):
        super(BinjaArtifactLifter, self).__init__(controller)

    def lift_addr(self, addr: int) -> int:
        return addr

    def lift_type(self, type_str: str) -> str:
        for bn_t, bs_t in self.lift_map.items():
            type_str = type_str.replace(bn_t, bs_t)

        return type_str

    def lift_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset

    def lower_addr(self, addr: int) -> int:
        return addr

    def lower_type(self, type_str: str) -> str:
        return type_str

    def lower_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset



import threading
import re

from PySide6.QtWidgets import QVBoxLayout
from PySide6.QtCore import Qt
from binaryninjaui import (
    UIContext,
    DockHandler,
    DockContextHandler,
    UIAction,
    UIActionHandler,
    Menu,
)
import binaryninja
from binaryninja import PluginCommand, BinaryView, SymbolType
from binaryninja.interaction import show_message_box
from binaryninja.enums import MessageBoxButtonSet, MessageBoxIcon, VariableSourceType
from binaryninja.binaryview import BinaryDataNotification

from collections import defaultdict
import logging

from binsync.common.ui.version import set_ui_version
set_ui_version("PySide6")
import binsync
from binsync.common.ui.config_dialog import SyncConfig
from binsync.common.ui.control_panel import ControlPanel
from .ui_tools import find_main_window, BinjaDockWidget, create_widget
from .controller import BinjaBinSyncController
from copy import deepcopy
from binsync.data import (
    State, User, Artifact,
    Function, FunctionHeader, FunctionArgument, StackVariable,
    Comment, GlobalVariable, Patch,
    Enum, Struct
)

l = logging.getLogger(__name__)

#
# Binja UI
#

class ControlPanelDockWidget(BinjaDockWidget):
    def __init__(self, controller, parent=None, name=None, data=None):
        super().__init__(name, parent=parent)
        self.data = data
        print(data)
        self._widget = None
        self.controller = controller

        self._init_widgets()

    def _init_widgets(self):
        self._widget = ControlPanel(self.controller)

        layout = QVBoxLayout()
        layout.addWidget(self._widget)
        self.setLayout(layout)


#
# Other
#

def instance():
    main_window = find_main_window()
    try:
        dock = [x for x in main_window.children() if isinstance(x, BinjaDockWidget)][0]
    except:
        dock = BinjaDockWidget("dummy")
    return dock


def conv_func_binja_to_binsync(binja_func):

    #
    # header: name, ret type, args
    #
    
    args = {
        i: FunctionArgument(i, parameter.name, parameter.type.get_string_before_name(), parameter.type.width)
        for i, parameter in enumerate(binja_func.parameter_vars)
    }

    sync_header = FunctionHeader(
        binja_func.name,
        binja_func.start,
        ret_type=binja_func.return_type.get_string_before_name(),
        args=args
    )

    #
    # stack vars
    #

    binja_stack_vars = {
        v.storage: v for v in binja_func.stack_layout if v.source_type == VariableSourceType.StackVariableSourceType
    }
    sorted_stack = sorted(binja_func.stack_layout, key=lambda x: x.storage)
    var_sizes = {}

    for off, var in binja_stack_vars.items():
        i = sorted_stack.index(var)
        if i + 1 >= len(sorted_stack):
            var_sizes[var] = 0
        else:
            var_sizes[var] = var.storage - sorted_stack[i].storage

    bs_stack_vars = {
        off: binsync.StackVariable(
            off, binsync.StackOffsetType.BINJA,
            var.name,
            var.type.get_string_before_name(),
            var_sizes[var],
            binja_func.start
        )
        for off, var in binja_stack_vars.items()
    }

    size = binja_func.address_ranges[0].end - binja_func.address_ranges[0].start
    return Function(binja_func.start, size, header=sync_header, stack_vars=bs_stack_vars)


class FunctionNotification(BinaryDataNotification):
    def __init__(self, view, controller):
        super().__init__()
        self._view = view
        self._controller = controller
        self._function_requested = None
        self._function_saved = None

    def function_updated(self, view, func_):
        if self._controller.sync_lock.locked():
            return

        # Service requested function only
        if self._function_requested == func_.start:
            #print(f"[BinSync] Servicing function: {func_.start:#x}")
            # Found function clear request
            self._function_requested = None

            # Convert to binsync Function type for diffing
            func = view.get_function_at(func_.start)
            bs_func = conv_func_binja_to_binsync(func)

            #
            # header
            #
            # note: function name done inside symbol update hook
            #

            # Check return type
            if self._function_saved.header.ret_type != bs_func.header.ret_type:
                self._controller.schedule_job(
                    self._controller.push_artifact,
                    bs_func.header
                )
                
            # Check arguments
            arg_changed = False
            for key, old_arg in self._function_saved.header.args.items():
                try:
                    new_arg = bs_func.header.args[key]
                except KeyError:
                    arg_changed = True
                    break

                if old_arg.name != new_arg.name:
                    arg_changed = True
                    break

                if old_arg.type_str != new_arg.type_str:
                    arg_changed = True
                    break

            if arg_changed:
                self._controller.schedule_job(
                    self._controller.push_artifact,
                    bs_func.header
                )

            #
            # stack vars
            #

            for off, var in self._function_saved.stack_vars.items():
                if off in bs_func.stack_vars and var != bs_func.stack_vars[off]:
                    new_var = bs_func.stack_vars[off]
                    if re.match(r"var_\d+[_\d+]{0,1}", new_var.name) \
                            or new_var.name in {'__saved_rbp', '__return_addr',}:
                        continue

                    self._controller.schedule_job(
                        self._controller.push_artifact,
                        new_var
                    )

            self._function_saved = None

    def function_update_requested(self, view, func):
        if not self._controller.sync_lock.locked() and self._function_requested is None:
            #print(f"[BinSync] Function requested {func.start:#x}")
            self._function_requested = func.start
            self._function_saved = conv_func_binja_to_binsync(func)
    
    def symbol_updated(self, view, sym):
        if self._controller.sync_lock.locked():
            return

        if sym.type == SymbolType.FunctionSymbol:
            func = view.get_function_at(sym.address)
            bs_func = conv_func_binja_to_binsync(func)
            self._controller.schedule_job(
                self._controller.push_artifact,
                FunctionHeader(sym.name, sym.address, ret_type=bs_func.header.ret_type, args=bs_func.header.args)
            )

        elif sym.type == SymbolType.DataSymbol:
            pass


class DataNotification(BinaryDataNotification):
    def __init__(self, view, controller):
        super().__init__()
        self._view = view
        self._controller = controller
        self._function_requested = None

    def data_var_updated(self, view, var):
        print(f"[BinSync] Data Updated Var: {var}, Type: {type(var)}")


class StructNotification(BinaryDataNotification):
    def __init__(self, view, controller):
        super().__init__()
        self._view = view
        self._controller = controller

    def type_defined(self, view:'BinaryView', name:'_types.QualifiedName', type:'_types.Type') -> None:
        print(f"[BinSync] Type defined!")

    def type_undefined(self, view:'BinaryView', name:'_types.QualifiedName', type:'_types.Type') -> None:
        print(f"[BinSync] Type undefined!")


def start_function_monitor(view, controller):
    notification = FunctionNotification(view, controller)
    view.register_notification(notification)

def start_data_monitor(view, controller):
    notification = DataNotification(view, controller)
    view.register_notification(notification)

def start_struct_monitor(view, controller):
    notification = StructNotification(view, controller)
    view.register_notification(notification)


class BinjaPlugin:
    def __init__(self):
        # controller stored by a binary view
        self.controllers = defaultdict(BinjaBinSyncController)
        self._init_ui()

    def _init_ui(self):
        # config dialog
        configure_binsync_id = "BinSync: Configure"
        UIAction.registerAction(configure_binsync_id)
        UIActionHandler.globalActions().bindAction(
            configure_binsync_id, UIAction(self._launch_config)
        )
        Menu.mainMenu("Tools").addAction(configure_binsync_id, "BinSync")

        # control panel (per BV)
        dock_handler = DockHandler.getActiveDockHandler()
        dock_handler.addDockWidget(
            "BinSync: Control Panel",
            lambda n, p, d: create_widget(ControlPanelDockWidget, n, p, d, self.controllers),
            Qt.RightDockWidgetArea,
            Qt.Vertical,
            True
        )

    def _init_bv_dependencies(self, bv):
        l.debug(f"Starting function hook")
        start_function_monitor(bv, self.controllers[bv])
        #Creates to much noise removing for time being
        #print(f"[BinSync] Starting data hook")
        #start_data_monitor(bv, self.controllers[bv])
        #print(f"[BinSync Starting struct hook")
        #start_struct_monitor(bv, self.controllers[bv])

    def _launch_config(self, bn_context):
        bv = bn_context.binaryView
        controller_bv = self.controllers[bv]

        # exit early if we already configed
        if controller_bv.bv is not None or bv is None:
            return
        controller_bv.bv = bv

        # configure
        dialog = SyncConfig(controller_bv)
        dialog.exec_()

        # if the config was successful init a full client
        if controller_bv.check_client():
            self._init_bv_dependencies(bv)


BinjaPlugin()

"""
PluginCommand.register(
    "Start Sharing Patches", "Start Sharing Patches", start_patch_monitor
)
PluginCommand.register(
    "Start Sharing Functions", "Start Sharing Functions", start_function_monitor
)
"""



import pathlib
import os
import toml

BINSYNC_CONFIG_POSTFIX = "bsconf"


class Config:
    __slots__ = (
        "path",
    )

    def __init__(self, path):
        self.path = path

    def save(self):
        if isinstance(self.path, str):
            path = pathlib.Path(self.path)

        if not path.parent.exists():
            return None

        dump_dict = {
            attr: getattr(self, attr) for attr in self.__slots__
        }

        with open(self.path, "w") as fp:
            toml.dump(dump_dict, fp)

        return self.path

    def load(self):
        path = pathlib.Path(self.path)
        if not path.exists():
            return None

        with open(self.path, "r") as fp:
            load_dict = toml.load(fp)

        for attr in self.__slots__:
            setattr(self, attr, load_dict.get(attr, None))

        return self

    @classmethod
    def load_from_file(cls, path):
        conf = cls(path)
        return conf.load()


class ProjectConfig(Config):
    __slots__ = Config.__slots__ + (
        "binary_name",
        "user",
        "repo_path",
        "remote",
    )

    def __init__(self,
                 binary_path,
                 user=None,
                 repo_path=None,
                 remote=None
                 ):
        super(ProjectConfig, self).__init__(self._correct_path(binary_path))

        self.binary_name = pathlib.Path(binary_path).name
        self.user = user
        self.repo_path = repo_path
        self.remote = remote

    def _correct_path(self, binary_path):
        # example config: /path/to/fauxware_files/.fauxware.bsconf
        binary_path = pathlib.Path(binary_path)
        config_name = pathlib.Path(f".{binary_path.name}.{BINSYNC_CONFIG_POSTFIX}")
        config_dir = binary_path.parent
        return str(config_dir.joinpath(config_name))


class GlobalConfig(Config):
    __slots__ = Config.__slots__ + (
        "last_bs_repo_path",
    )

    def __init__(self,
                 path,
                 last_bs_repo_path=None
                 ):
        super(GlobalConfig, self).__init__(self._correct_path(path))

        self.last_bs_repo_path = last_bs_repo_path

    def _correct_path(self, path):
        if path is None:
            path = os.getenv("HOME") or "."

        path = pathlib.Path(path)
        name = pathlib.Path(f".global.{BINSYNC_CONFIG_POSTFIX}")
        return str(path.joinpath(name))
